#!/usr/bin/env python
# coding: utf-8

# # Modeling and forecasting the early evolution of the Covid-19 pandemic in Brazil
# 
# Saulo B. Bastos, Daniel O. Cajueiro
# 
# https://arxiv.org/abs/2003.14288

# In[1]:


import os
import sys
import argparse
import numpy as np
import pandas as pd
from datetime import datetime, date, timedelta
import collections
import pickle
from functools import partial
from scipy.integrate import solve_ivp
from scipy.interpolate import interp1d
import scipy.optimize
import matplotlib
import matplotlib.pyplot as plt

#from IPython.display import IFrame, HTML, Markdown, display
from tabulate import tabulate

# ***************************************************************************************************************************************

parser = argparse.ArgumentParser()

parser.add_argument("--least_squares_loss", default="cauchy")
parser.add_argument("--random_seed", type=int, default=7)
parser.add_argument("--least_squares_f_scale", type=float, default=2.0)

args = parser.parse_args()

# ***************************************************************************************************************************************

def is_computer_using_windows():
    return (sys.platform.lower().startswith('win') or (os.name=='nt'))

def path_join(*paths):
    return os.path.join(*paths) if (not is_computer_using_windows()) else os.path.join(*paths).replace("/", "\\")

dir_separator = "\\" if is_computer_using_windows() else "/"

data_dir = path_join(dir_separator.join(os.getcwd().split(dir_separator)[:-1]), "data"+dir_separator)
if (not os.path.isdir(data_dir)):
    os.makedirs(data_dir)
print("Data directory: ", data_dir)

base_output_dir = path_join(dir_separator.join(os.getcwd().split(dir_separator)[:-1]), "output"+dir_separator)
if (not os.path.isdir(base_output_dir)):
    os.makedirs(base_output_dir)

output_dir = path_join(base_output_dir, "rand-"+str(args.random_seed)+dir_separator) if np.isclose(args.least_squares_f_scale, 1.0) else path_join(base_output_dir, "rand-"+str(args.random_seed)+"_fscale-"+str(args.least_squares_f_scale)+dir_separator)
if (not os.path.isdir(output_dir)):
    os.makedirs(output_dir)
print("Output directory: ", output_dir)

current_date = date(2020, 3, 30)
print("Last date: ", current_date)

if (not os.path.exists(os.path.join(data_dir, "brazil_states.csv"))):
    brazil_states_df = pd.DataFrame(data=[["AC",12], ["AL",27], ["AM",13], ["AP",16], ["BA",29], ["CE",23], ["DF",53], ["ES",32], ["GO",52], ["MA",21], ["MG",31], ["MS",50], ["MT",51], ["PA",15], ["PB",25], ["PE",26], ["PI",22], ["PR",41], ["RJ",33], ["RN",24], ["RO",11], ["RR",14], ["RS",43], ["SC",42], ["SE",28], ["SP",35], ["TO",17]], columns=["uf","uid"])

    regions_dict = {"norte": ["AC", "AP", "AM", "PA", "RR", "RO", "TO"],
                    "nordeste": ["AL", "BA", "CE", "MA", "PB", "PE", "PI", "RN", "SE"],
                    "centro-oeste": ["GO", "MT", "MS"],
                    "sudeste": ["SP", "RJ", "MG", "ES"],
                    "sul": ["PR", "SC", "RS"],
                    "distrito-federal": ["DF"]}
    brazil_states_df["region"] = brazil_states_df["uf"].apply(lambda uf: next(region for region, uf_list in regions_dict.items() if (uf in uf_list)))

    brazil_states_df.to_csv(os.path.join(data_dir, "brazil_states.csv"), sep=";", header = True, mode = 'w', index=False)
else:
    brazil_states_df = pd.read_csv(os.path.join(data_dir, "brazil_states.csv"), sep=";")

state_being_analysed = "BR"
simulation_csv = path_join(output_dir, "simulations-"+state_being_analysed+"_loss-"+args.least_squares_loss+".csv")

date_of_effectiveness_of_policy = date(2020, 3, 22)
policy_begin_date = date_of_effectiveness_of_policy + timedelta(days=1)

use_median_for_parameters = False
random_seed = args.random_seed

confidence_interval = 0.05
f_agg = [np.mean, partial(np.quantile, q=0.5)][0]

generate_figures = True

# # Function definitions

# In[6]:


def diff_SIR(t,y,beta,gamma,mu=0.0,Lambda=0.0,rho=0.0,N=None,transmition_factor_fn=None):
    psi = transmition_factor_fn(t) if (not transmition_factor_fn is None) else 1.0

    S, I, R, D = y[0], y[1], y[2], y[3]
    _N = N-D
    #_N = N if (not isinstance(N, collections.Iterable)) and np.isclose(Lambda, mu*N) else interp1d(t_train, N_train, kind="previous", fill_value="extrapolate")(t)
    
    dS = Lambda - mu*S - psi*beta*I*S/_N
    dI = psi*beta*I*S/_N - (gamma + mu)*I / (1-rho)
    dR = gamma*I - mu*R
    dD = rho * (gamma + mu) * I / (1-rho)
    return np.array([dS, dI, dR, dD])

def diff_SIAS(t,y,beta_A,beta_S,gamma_A,gamma_S,mu=0.0,Lambda=0.0,rho=0.0,delta=0.0,N=None,t_train=None,N_train=None,beta_factor_fn=None):
    S, I_A, I_S, D = y[0], y[1], y[2], y[3]
    _N = N-D
    #_N = N if (not isinstance(N, collections.Iterable)) and np.isclose(Lambda, mu*N) else interp1d(t_train, N_train, kind="previous", fill_value="extrapolate")(t)

    
    dS = Lambda - mu*S - (beta_A*I_A + beta_S*I_S)*S/_N
    dI_A = (beta_A*I_A + beta_S*I_S)*S/_N - (delta + gamma_A + mu)*I_A
    dI_S = delta*I_A - (gamma_S + mu)*I_S / (1-rho)
    dD = rho * (gamma_S + mu) * I_S / (1-rho)
    return np.array([dS, dI_A, dI_S, dD])

def diff_SIAS_with_p(t,y,beta_A,beta_S,gamma_A,gamma_S,mu=0.0,Lambda=0.0,rho=0.0,p=0.0,N=None,transmition_factor_fn=None):
    psi = transmition_factor_fn(t) if (not transmition_factor_fn is None) else 1.0

    S, I_A, I_S, R_A, R_S, D = y

    #_N = N if (not isinstance(N, collections.Iterable)) and np.isclose(Lambda, mu*N) else interp1d(t_train, N_train, kind="previous", fill_value="extrapolate")(t)
    _N = N-D

    dS = Lambda - mu*S - psi*(beta_A*I_A + beta_S*I_S)*S/_N
    dI_A = (1-p)*psi*(beta_A*I_A + beta_S*I_S)*S/_N - (gamma_A + mu)*I_A
    dI_S = p*psi*(beta_A*I_A + beta_S*I_S)*S/_N - (gamma_S + mu)*I_S / (1-rho)
    dR_A = gamma_A*I_A - mu*R_A
    dR_S = gamma_S*I_S - mu*R_S
    dD = rho*(gamma_S + mu)*I_S/(1-rho)
    return np.array([dS, dI_A, dI_S, dR_A, dR_S, dD])

Lambda = 0
mu = 0

def t2date(t, pos=None):
    first_date = data_df.iloc[0]["date"]
    return first_date if np.isclose(t, 0) else first_date + timedelta(days=int(t))

import matplotlib.ticker as ticker
def plot_SIR(ax, t, S, I, R=None, D=None):
    line_S, = ax.plot(t, S, label="S", color="blue")
    line_I, = ax.plot(t, I, label="I", color="orange")
    if (not R is None):
        line_R, = ax.plot(t, R, label="R", color="green")

    if (not D is None):
        line_D, = ax.plot(t, D, label="D", color="red")
    ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
    ax.set_ylabel("# of people")

    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    ax.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)

    return

def plot_infected_and_deaths(t, I, D):
    fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True)
    line_I, = axs[0].plot(t, I, label="I", color="orange")
    axs[0].set_ylabel("# infected")
    axs[0].yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))

    line_D, = axs[1].plot(t, D, label="D", color="red")
    axs[1].set_ylabel("# deads")
    
    axs[1].xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    axs[1].yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))
    plt.setp(axs[1].xaxis.get_majorticklabels(), rotation=45)
    
    plt.show()

    return

def cont2disc(t, x, x2=None):
    from scipy.interpolate import interp1d
    t_disc = np.arange(start=np.round(min(t)), stop=np.round(max(t))+1)

    f = interp1d(t, x, kind="linear", fill_value="extrapolate")
    if (x2 is None):
        return t_disc, f(t_disc)
    else:
        f2 = interp1d(t, x2, kind="linear", fill_value="extrapolate")
        return t_disc, f(t_disc), f2(t_disc)
        

def plot_SIR_prediction(ax, data_df, variable, t, x, num_steps_ahead=3, prediction_start_date=None):
    line_I_real, = ax.plot(data_df["t"], data_df[variable], label="Real", color="DarkOrange", marker="*", linestyle="")
    
    t_indexes = [i for i, ti in enumerate(t) if (ti < data_df["t"].max()+num_steps_ahead)]
    line_I, = ax.plot(t[t_indexes], x[t_indexes], label="Simulation", color="orange")

    t_prediction_start = date2t(prediction_start_date) if (not prediction_start_date is None) else data_df["t"].max()
    
    t_points, I_points = cont2disc(t, x)
    t_points, I_points = zip(*[(ti, I_points[i]) for i, ti in enumerate(t_points) if (ti > t_prediction_start) and (ti <= max(t[t_indexes]))])
    line_I_points, = ax.plot(t_points, I_points, label="Simulation prediction", color="Chocolate", marker=".", linestyle="")

    ax.plot([t_prediction_start, t_prediction_start], [0, max(x)], color="gray", linestyle="--")

    ax.set_xlim(data_df["t"].min(), max(t[t_indexes]))
    ax.set_ylim(0, max(max(x[t_indexes]), max(data_df[variable])))
    ax.set_ylabel(variable)

    ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)

    return

def get_SIR_statistics(t, I):
    SIR_statistics = dict()
    
    N = data_df.iloc[0]["N"]

    t_cont_at_peak = t[np.argmax(I)]
    SIR_statistics["peak_date"] = data_df.iloc[0]["date"] + timedelta(days=np.round(t_cont_at_peak))
    SIR_statistics["peak_percent"] = np.max(I)/N

    return SIR_statistics

def get_SIR_prediction_table(t, I, num_steps_ahead=3):
    prediction_df = pd.DataFrame(columns = ["t", "date", "real", "prediction"])
    t_points, I_points = cont2disc(t, I)
    
    for k, row in data_df.iterrows():
        prediction = next((I_points[i] for i, ti in enumerate(t_points) if np.isclose(ti, row["t"])), None)
        prediction_df.loc[len(prediction_df)] = [row["t"], row["date"], row["I"], np.round(prediction) if (not prediction is None) else None]

    last_t = data_df["t"].max()
    last_date = data_df["date"].max()
    for step in range(num_steps_ahead):
        prediction = next((I_points[i] for i, ti in enumerate(t_points) if np.isclose(ti, last_t+step+1)), None)
        prediction_df.loc[len(prediction_df)] = [last_t+step+1, last_date+timedelta(days=step+1), None, np.round(prediction) if (not prediction is None) else None]

    return prediction_df

def get_transmition_factor_fn(psi,date_i,date_f=None):
    if (date_f is None):
        return lambda t: psi if (t>date2t(date_i)) else 1.0
    else:
        return lambda t: psi if (t>date2t(date_i)) and (t<=date2t(date_f)) else 1.0

def get_is_policy_effective_fn(date_i,date_f=None):
    if (date_f is None):
        return lambda t: (t>date2t(date_i))
    else:
        return lambda t: (t>date2t(date_i)) and (t<=date2t(date_f))

def get_SIR_params_bounds(): 
    return {"beta": (1/10, 1/0.5), "gamma": (1/14, 1/2), "rho": (0.01, 0.10), "psi": (0, 1)}

def get_SIR_parameters_from_optimization(data_df, beta=None, gamma=1/14, mu=0.0, Lambda=0.0, rho=0.0, is_policy_effective_fn = None, variables_to_optimize = ["I", "D"], num_iterations = 100, t0=1, x0 = None, return_params_list = False, plot_params_histogram = False):
    params_to_optimize = list()
    params_bounds_l = list()
    params_bounds_u = list()
    
    sir_params_bounds = get_SIR_params_bounds()
    if (beta is None):
        params_to_optimize.append("beta")
        params_bounds_l.append(sir_params_bounds["beta"][0])
        params_bounds_u.append(sir_params_bounds["beta"][1])

    if (gamma is None):
        params_to_optimize.append("gamma")
        params_bounds_l.append(sir_params_bounds["gamma"][0])
        params_bounds_u.append(sir_params_bounds["gamma"][1])

    if (rho is None):
        params_to_optimize.append("rho")
        params_bounds_l.append(sir_params_bounds["rho"][0])
        params_bounds_u.append(sir_params_bounds["rho"][1])

    if (not is_policy_effective_fn is None):
        params_to_optimize.append("psi")
        params_bounds_l.append(sir_params_bounds["psi"][0])
        params_bounds_u.append(sir_params_bounds["psi"][1])

    def loss_sir_model(params, t, x):
        I_new, I_acc, D_new, D_acc = x

        _beta = params[params_to_optimize.index("beta")] if ("beta" in params_to_optimize) else beta
        _gamma = params[params_to_optimize.index("gamma")] if ("gamma" in params_to_optimize) else gamma
        _rho = params[params_to_optimize.index("rho")] if ("rho" in params_to_optimize) else rho
        if ("psi" in params_to_optimize):
            _transmition_factor_fn = lambda t: params[params_to_optimize.index("psi")] if is_policy_effective_fn(t) else 1.0 
        else:
            _transmition_factor_fn = None

        integration_result = solve_ivp(fun=partial(diff_SIR,beta=_beta, gamma=_gamma, mu=mu, Lambda=Lambda, rho=_rho, N=data_df.iloc[t0]["N"], transmition_factor_fn=_transmition_factor_fn), t_span=(1, data_df["t"].max()), y0=[S0, I0, 0, 1], method="RK45", max_step=0.1)
        t_simulation = integration_result.t
        S_simulation, I_simulation, R_simulation, D_simulation = integration_result.y

        t_in_evaluation = t if (_transmition_factor_fn is None) else [ti for ti in t if is_policy_effective_fn(ti)]
        I_approx = interp1d(t_simulation, I_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)
        R_approx = interp1d(t_simulation, R_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)
        D_approx = interp1d(t_simulation, D_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)

        if (len(t_in_evaluation) < len(t)):
            t_in_evaluation_indexes = [np.where(t==ti)[0][0] for ti in t_in_evaluation]
            I_new, I_acc, D_new, D_acc = I_new[t_in_evaluation_indexes], I_acc[t_in_evaluation_indexes], D_new[t_in_evaluation_indexes], D_acc[t_in_evaluation_indexes]

        x_real, x_approx = list(), list()
        if ("I" in variables_to_optimize):
            x_real.append(I_acc-D_acc)
            x_approx.append(I_approx+R_approx)

        if ("D" in variables_to_optimize):
            x_real.append(D_acc)
            x_approx.append(D_approx)

        if (len(x_approx) > 0):
            #return np.abs(np.concatenate(np.log(1+np.array(x_approx))) - np.concatenate(np.log(1+np.array(x_real))))
            return np.abs(np.concatenate(x_approx) - np.concatenate(x_real))
        else:
            raise Exception("Invalid cost!")

    t_train, y_train = data_df["t"].values[t0:], (data_df["cases_new"].values[t0:], data_df["cases"].values[t0:], data_df["deaths_new"].values[t0:], data_df["deaths"].values[t0:])

    params = dict()
    for p in params_to_optimize:
        params[p] = list()

    for iteration in range(num_iterations):
        print("iteration = ", iteration+1, "/", num_iterations, end="", flush=True)
        x0_values = list()
        for i_param, param_name in enumerate(params_to_optimize):
            if (not x0 is None) and (param_name in x0):
                suggested_value = x0[param_name]
                delta_i = params_bounds_u[i_param] - params_bounds_l[i_param]
                if (iteration > 0):
                    suggested_value = suggested_value + delta_i*(np.random.random()-0.5)/10.0
                    suggested_value = min(max(suggested_value, params_bounds_l[i_param]), params_bounds_u[i_param])
                x0_values.append(suggested_value)
            elif ("gamma" in param_name):
                mean_i = np.exp((np.log(params_bounds_l[i_param])+np.log(params_bounds_u[i_param]))/2)
                std_i = np.exp((np.log(params_bounds_l[i_param])+np.log(params_bounds_u[i_param])))
                guess = np.random.normal(mean_i, std_i)
                guess = min(max(guess, params_bounds_l[i_param]), params_bounds_u[i_param])
                x0_values.append(guess)
            else:
                value_i = params_bounds_l[i_param]
                delta_i = params_bounds_u[i_param] - params_bounds_l[i_param]
                guess = value_i + delta_i*np.random.random()
                x0_values.append(guess)

        if (np.isclose(args.least_squares_f_scale, 1.0)):
            res_lsq = scipy.optimize.least_squares(loss_sir_model, x0=x0_values, loss=args.least_squares_loss, jac='3-point', method="trf", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-10, tr_solver="exact", verbose=0) # , method="lm"
        else:
            res_lsq = scipy.optimize.least_squares(loss_sir_model, x0=x0_values, loss=args.least_squares_loss, f_scale=args.least_squares_f_scale, jac='3-point', method="trf", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-10, tr_solver="exact", verbose=0) # , method="lm"
            
        #res_lsq = scipy.optimize.least_squares(loss_sir_model, x0=x0_values, loss="cauchy", jac='3-point', method="trf", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-9, xtol=None, tr_solver="exact", verbose=0) # , method="lm"

        print(" -> (", end="", flush=True)
        for i_param, p in enumerate(params_to_optimize):
            print(" " + p + "=" + str(res_lsq.x[i_param]), end="", flush=True)
        print(")", end="", flush=True)

        simulation_hasnt_hit_bounds = True
        for i_param, p in enumerate(params_to_optimize):
            if (np.isclose(params_bounds_l[i_param], res_lsq.x[i_param]) or np.isclose(params_bounds_u[i_param], res_lsq.x[i_param])):
                simulation_hasnt_hit_bounds = False
                break

        print(" -> simulation_has_hit_bounds = " + str(int(not simulation_hasnt_hit_bounds)), end="")
        if (not simulation_hasnt_hit_bounds):
            print(" -> ignored")
            continue
        print("")

        for i_param, p in enumerate(params_to_optimize):
            params[p].append(res_lsq.x[i_param])

    if (return_params_list):
        import copy
        params_list = copy.deepcopy(params)

    for key in params.keys():
        if (plot_params_histogram):
            fig, ax = plt.subplots(nrows=1, ncols=1)
            n, bins, patches = ax.hist(x=params[key], bins='auto', color='#0504aa', alpha=0.7, rwidth=0.85)
            ax.set_xlabel('Value')
            ax.set_ylabel('Frequency')
            ax.set_title(key)
            ax.set_xlim(min(params[key]), max(params[key]))
            plt.show()
        
        params[key] = np.mean(params[key])

    if (return_params_list):
        return params, params_list

    return params

def get_SIAS_with_p_parameters_from_optimization(data_df, beta_A=None, beta_S=None, gamma_A=1/14, gamma_S=1/14, mu=0.0, Lambda=0.0, rho=0.0, p = None, is_policy_effective_fn = None, force_same_beta = False, variables_to_optimize = ["I", "D"], p_IA0=0, t0=1, x0 = None, num_iterations = 100, return_params_list = False, plot_params_histogram = False):

    if (isinstance(variables_to_optimize, str)):
        variables_to_optimize = [variables_to_optimize]
    
    params_to_optimize = list()
    params_bounds_l = list()
    params_bounds_u = list()
    if (beta_A is None) or (beta_S is None):
        if (force_same_beta):
            params_to_optimize.append("beta")
            params_bounds_l.append(1/10)
            params_bounds_u.append(1)
        elif (beta_A is None) and (not beta_S is None):
            params_to_optimize.append("beta_A")
            params_bounds_l.append(1/10)
            params_bounds_u.append(beta_S)
        elif (not beta_A is None) and (beta_S is None):
            params_to_optimize.append("beta_S")
            params_bounds_l.append(beta_A)
            params_bounds_u.append(1)
        else:
            params_to_optimize.append("beta_A")
            params_bounds_l.append(1/10)
            params_bounds_u.append(1)

            params_to_optimize.append("c")
            params_bounds_l.append(1)
            params_bounds_u.append(20)

    if (gamma_A is None):
        params_to_optimize.append("gamma_A")
        params_bounds_l.append(1/14)
        params_bounds_u.append(1/2)

    if (gamma_S is None):
        params_to_optimize.append("gamma_S")
        params_bounds_l.append(1/14)
        params_bounds_u.append(1/2)

    if (rho is None):
        params_to_optimize.append("rho")
        params_bounds_l.append(0.01)
        params_bounds_u.append(0.10)

    if (p is None):
        params_to_optimize.append("p")
        params_bounds_l.append(0)
        params_bounds_u.append(1)

    if (not is_policy_effective_fn is None):
        params_to_optimize.append("psi")
        params_bounds_l.append(0.25)
        params_bounds_u.append(1)

    def loss_sias_model(params, t, x):
        I_new, I_acc, D_new, D_acc = x

        if ("beta" in params_to_optimize):
            beta = params[params_to_optimize.index("beta")]
            _beta_A, _beta_S = beta, beta
        elif ("beta_A" in params_to_optimize) and ("c" in params_to_optimize):
            _beta_A = params[params_to_optimize.index("beta_A")] if ("beta_A" in params_to_optimize) else beta_A
            _beta_S = params[params_to_optimize.index("c")]*_beta_A if ("c" in params_to_optimize) else beta_S
        elif ("beta_A" in params_to_optimize) and (not "c" in params_to_optimize):
            _beta_A = params[params_to_optimize.index("beta_A")] if ("beta_A" in params_to_optimize) else beta_A
            _beta_S = beta_S
        elif ("beta_S" in params_to_optimize) and (not "beta_A" in params_to_optimize) and (not "c" in params_to_optimize):
            _beta_A = beta_A
            _beta_S = params[params_to_optimize.index("beta_S")] if ("beta_S" in params_to_optimize) else beta_S
        else:
            _beta_A = beta_A
            _beta_S = beta_S

        _gamma_A = params[params_to_optimize.index("gamma_A")] if ("gamma_A" in params_to_optimize) else gamma_A
        _gamma_S = params[params_to_optimize.index("gamma_S")] if ("gamma_S" in params_to_optimize) else gamma_S
        _rho = params[params_to_optimize.index("rho")] if ("rho" in params_to_optimize) else rho
        _p = params[params_to_optimize.index("p")] if ("p" in params_to_optimize) else p

        if ("psi" in params_to_optimize):
            _transmition_factor_fn = lambda t: params[params_to_optimize.index("psi")] if is_policy_effective_fn(t) else 1.0
        else:
            _transmition_factor_fn = None

        integration_result = solve_ivp(fun=partial(diff_SIAS_with_p,beta_A=_beta_A,beta_S=_beta_S,gamma_A=_gamma_A,gamma_S=_gamma_S,mu=mu,Lambda=Lambda,rho=_rho,p=_p, N=data_df.iloc[t0]["N"], transmition_factor_fn=_transmition_factor_fn), t_span=(t0, data_df["t"].max()), y0=[S0-p_IA0, p_IA0, I0, 0, 0, 1], method="RK45", max_step=0.1)
        t_simulation, IS_simulation, RS_simulation, D_simulation = integration_result.t, integration_result.y[2], integration_result.y[4], integration_result.y[-1]

        t_in_evaluation = t if (_transmition_factor_fn is None) else [ti for ti in t if is_policy_effective_fn(ti)]
        RS_approx = interp1d(t_simulation, RS_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)
        IS_approx = interp1d(t_simulation, IS_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)
        D_approx = interp1d(t_simulation, D_simulation, kind="linear", fill_value="extrapolate")(t_in_evaluation)

        if (len(t_in_evaluation) < len(t)):
            t_in_evaluation_indexes = [np.where(t==ti)[0][0] for ti in t_in_evaluation]
            I_new, I_acc, D_new, D_acc = I_new[t_in_evaluation_indexes], I_acc[t_in_evaluation_indexes], D_new[t_in_evaluation_indexes], D_acc[t_in_evaluation_indexes]

        x_real, x_approx = list(), list()
        if ("I" in variables_to_optimize):
            x_real.append(I_acc-D_acc)
            x_approx.append(IS_approx+RS_approx)

        if ("D" in variables_to_optimize):
            x_real.append(D_acc)
            x_approx.append(D_approx)
            
        if (len(x_approx) > 0):
            return np.concatenate(x_approx) - np.concatenate(x_real)
        else:
            raise Exception("Invalid cost!")

    t_train, y_train = data_df["t"].values[t0:], (data_df["cases_new"].values[t0:], data_df["cases"].values[t0:], data_df["deaths_new"].values[t0:], data_df["deaths"].values[t0:])

    params = dict()
    for name in params_to_optimize:
        params[name] = list()

    for iteration in range(num_iterations):
        print("iteration = ", iteration+1, "/", num_iterations, end="", flush=True)

        x0_values = list()
        for i_param, param_name in enumerate(params_to_optimize):
            if (not x0 is None) and (param_name in x0):
                suggested_value = x0[param_name]
                delta_i = params_bounds_u[i_param] - params_bounds_l[i_param]
                if (iteration > 0):
                    suggested_value = suggested_value + delta_i*(np.random.random()-0.5)/10.0
                    suggested_value = min(max(suggested_value, params_bounds_l[i_param]), params_bounds_u[i_param])
                x0_values.append(suggested_value)
            elif ("gamma" in param_name):
                mean_i = np.exp((np.log(params_bounds_l[i_param])+np.log(params_bounds_u[i_param]))/2)
                std_i = np.exp((np.log(params_bounds_l[i_param])+np.log(params_bounds_u[i_param])))
                guess = np.random.normal(mean_i, std_i)
                guess = min(max(guess, params_bounds_l[i_param]), params_bounds_u[i_param])
                x0_values.append(guess)
            else:
                value_i = params_bounds_l[i_param]
                delta_i = params_bounds_u[i_param] - params_bounds_l[i_param]
                guess = value_i + delta_i*np.random.random()
                x0_values.append(guess)

        #res_lsq = scipy.optimize.least_squares(loss_sias_model, x0=x0_values, loss="linear", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-9, xtol=None, verbose=0) # , method="lm"
        if (np.isclose(args.least_squares_f_scale, 1.0)):
            res_lsq = scipy.optimize.least_squares(loss_sias_model, x0=x0_values, loss=args.least_squares_loss, jac='3-point', method="trf", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-10, tr_solver="exact", verbose=0) # , method="lm"
        else:
            res_lsq = scipy.optimize.least_squares(loss_sias_model, x0=x0_values, loss=args.least_squares_loss, f_scale=args.least_squares_f_scale, jac='3-point', method="trf", bounds=(params_bounds_l, params_bounds_u), args=(t_train, y_train), ftol=1e-10, tr_solver="exact", verbose=0) # , method="lm"

        print(" -> (", end="", flush=True)
        for i_param, param_name in enumerate(params_to_optimize):
            print(" " + param_name + "=" + str(res_lsq.x[i_param]), end="", flush=True)
        print(")", end="", flush=True)

        simulation_hasnt_hit_bounds = True
        for i_param, param_name in enumerate(params_to_optimize):
            if (param_name=="beta_A") and np.isclose(params_bounds_l[i_param], res_lsq.x[i_param]):
                simulation_hasnt_hit_bounds = False
                break
            elif (param_name!="beta_A") and ((np.isclose(params_bounds_l[i_param], res_lsq.x[i_param]) or np.isclose(params_bounds_u[i_param], res_lsq.x[i_param]))):
                simulation_hasnt_hit_bounds = False
                break

        print(" -> simulation_has_hit_bounds = " + str(int(not simulation_hasnt_hit_bounds)), end="")
        if (not simulation_hasnt_hit_bounds):
            print(" -> ignored")
            continue
        print("")

        for i_param, name in enumerate(params_to_optimize):
            params[name].append(res_lsq.x[i_param] if (name != "c") else res_lsq.x[i_param]*res_lsq.x[params_to_optimize.index("beta_A")])

    if ("c" in params_to_optimize):
        params["beta_S"] = params["c"]
        del params["c"]
    elif ("beta" in params_to_optimize):
        params["beta_A"] = params["beta"]
        params["beta_S"] = params["beta"]
        del params["beta"]

    if (return_params_list):
        import copy
        params_list = copy.deepcopy(params)

    for key in params.keys():
        if (plot_params_histogram):
            fig, ax = plt.subplots(nrows=1, ncols=1)
            n, bins, patches = ax.hist(x=params[key], bins='auto', color='#0504aa', alpha=0.7, rwidth=0.85)
            #plt.grid(axis='y', alpha=0.75)
            ax.set_xlabel('Value')
            ax.set_ylabel('Frequency')
            ax.set_title(key)
            ax.set_xlim(min(params[key]), max(params[key]))
            plt.show()
        
        params[key] = np.mean(params[key])

    if (return_params_list):
        return params, params_list

    return params

def date2t(d):
    return (d-data_df.iloc[0]["date"]).days

def get_SIR_description(cut_date, policy_begin_date=None):
    if (not policy_begin_date is None):
        return "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
    else:
        return "SIR without social distance (cut: " + cut_date.isoformat() + ")"
    
def get_AS_description(cut_date, p_IA0, policy_begin_date=None, rho=None):
    if (not policy_begin_date is None):
        return "Assymptomatic/symptomatic model with social distance, p_IA0="+str(p_IA0)+ " (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")" + ("" if (rho is None) else " - using Ministry of Health mortality of 2.8%")
    else:
        return "Assymptomatic/symptomatic model, p_IA0="+str(p_IA0)+ " (cut: " + cut_date.isoformat() + ")" + ("" if (rho is None) else " - using Ministry of Health mortality of 2.8%")
    
def get_quantile(simulations, key, param, q=0.5):
    return np.quantile(simulations[key]["params_list"][param], q)

def get_SIR_params(simulations, key, use_median = True):
    f_agg = partial(np.quantile, q=0.5) if (use_median) else np.mean
    return {"beta": f_agg(simulations[key]["params_list"]["beta"]),
            "gamma": f_agg(simulations[key]["params_list"]["gamma"]),
            "rho": f_agg(simulations[key]["params_list"]["rho"])}

def simulate_SIR(simulations, key, df, use_median = True, n_steps_ahead = 150, policy_date_i = None, policy_date_f = None):
    params = get_SIR_params(simulations=simulations, key=key, use_median=use_median)
    if ("psi" in simulations[key]["params"]) and (not simulations[key]["params"]["psi"] is None):
        if ("psi" in simulations[key]["params_list"]):
            f_agg = partial(np.quantile, q=0.5) if (use_median) else np.mean
            _psi = f_agg(simulations[key]["params_list"]["psi"])
        else:
            _psi = simulations[key]["params"]["psi"]

        date_i = policy_date_i if (not policy_date_i is None) else date_of_effectiveness_of_policy
        transmition_factor_fn = get_transmition_factor_fn(psi=_psi, date_i=date_i, date_f=policy_date_f)
    else:
        transmition_factor_fn = None

    results = solve_ivp(fun=partial(diff_SIR, **params, N=df.iloc[0]["N"], transmition_factor_fn=transmition_factor_fn), t_span=(1, data_df["t"].max()+n_steps_ahead), y0=[S0, I0, 0, 1], method="RK45", max_step=0.1)
    return results

def get_SIAS_params(simulations, sir_key, sias_key, psi=None, use_median = True, policy_date_i = None, policy_date_f = None):
    f_agg = partial(np.quantile, q=0.5) if (use_median) else np.mean

    params = {"beta_S": f_agg(simulations[sir_key]["params_list"]["beta"]),
              "gamma_S": f_agg(simulations[sir_key]["params_list"]["gamma"]),
              "rho": f_agg(simulations[sir_key]["params_list"]["rho"]),
              "beta_A": f_agg(simulations[sias_key]["params_list"]["beta_A"]),
              "gamma_A": f_agg(simulations[sias_key]["params_list"]["gamma_A"]),
              "p": f_agg(simulations[sias_key]["params_list"]["p"])}
    if (not psi is None):
        date_i = policy_date_i if (not policy_date_i is None) else date_of_effectiveness_of_policy
        params["transmition_factor_fn"] = get_transmition_factor_fn(psi=psi, date_i=date_i, date_f=policy_date_f)

    return params

def get_bounds_SIR(data_df, simulations, simulation_key, parent_key=None, policy_date_i=None, policy_date_f=None, use_fast_interval_bounds=False):
    I_simulations = {"I": list(), "R": list(), "D": list()}

    if (parent_key is None):
        parent_key = simulation_key

    confidence_params = dict()
    params_intervals_to_limit_simulations = dict()
    num_intervals = 10 if (use_fast_interval_bounds) else None
    for param in ["beta", "gamma", "rho"]:
        confidence_params[param] = (np.quantile(simulations[parent_key]["params_list"][param], confidence_interval/2), np.quantile(simulations[parent_key]["params_list"][param], 1-confidence_interval/2))

        if (not num_intervals is None):
            intervals = np.linspace(start=confidence_params[param][0], stop=confidence_params[param][1]+100*np.finfo(float).eps, num=num_intervals+1)
            params_intervals_to_limit_simulations[param] = {"interval": list(), "used_in_interval_generation": list()}
            for k in range(num_intervals):
                params_intervals_to_limit_simulations[param]["interval"].append( (intervals[k], intervals[k+1]) )
                params_intervals_to_limit_simulations[param]["used_in_interval_generation"].append(False)

    t_bounds = None
    for beta_i, gamma_i, rho_i in zip(simulations[parent_key]["params_list"]["beta"], simulations[parent_key]["params_list"]["gamma"], simulations[parent_key]["params_list"]["rho"]):
        
        all_params_are_ok = True
        for param, param_value in [("beta", beta_i), ("gamma", gamma_i), ("rho", rho_i)]:
            all_params_are_ok = all_params_are_ok and (param_value > confidence_params[param][0]) and (param_value < confidence_params[param][1])

        if (not all_params_are_ok):
            continue

        for param, param_value in [("beta", beta_i), ("gamma", gamma_i), ("rho", rho_i)]:
            can_be_used_in_interval_generation = False
            
            if (param in params_intervals_to_limit_simulations):
                for k in range(num_intervals):
                    if ((param_value >= params_intervals_to_limit_simulations[param]["interval"][k][0]) and (param_value <= params_intervals_to_limit_simulations[param]["interval"][k][1])):
                        if (not params_intervals_to_limit_simulations[param]["used_in_interval_generation"][k]):
                            can_be_used_in_interval_generation = True
                        params_intervals_to_limit_simulations[param]["used_in_interval_generation"][k] = True
            else:
                can_be_used_in_interval_generation = True

        if (not can_be_used_in_interval_generation):
            continue

        if (simulation_key != parent_key) and ("psi" in simulations[simulation_key]["params"]):
            psi_i = simulations[simulation_key]["params"]["psi"]
            date_i = policy_date_i if (not policy_date_i is None) else date_of_effectiveness_of_policy
            transmition_factor_fn = get_transmition_factor_fn(psi=psi_i, date_i=date_i, date_f=policy_date_f)
        else:
            transmition_factor_fn = None

        t0 = 1
        results = solve_ivp(fun=partial(diff_SIR, beta=beta_i, gamma=gamma_i, rho=rho_i, Lambda=simulations[parent_key]["params"]["Lambda"], mu=simulations[parent_key]["params"]["mu"], N=df.iloc[t0]["N"], transmition_factor_fn=transmition_factor_fn), t_span=(t0, data_df["t"].max()+150), y0=[S0, I0, 0, 1], method="RK45", max_step=0.1)
        
        if (t_bounds is None):
            t_bounds = results.t

        if (len(results.t) == len(t_bounds)):
            I_simulations["I"].append(list(results.y[1]))
            I_simulations["R"].append(list(results.y[2]))
            I_simulations["D"].append(list(results.y[3]))
        else:
            f_I = interp1d(results.t, results.y[1], kind="linear", fill_value="extrapolate")
            I_simulations["I"].append(list(f_I(t_bounds)))

            f_R = interp1d(results.t, results.y[2], kind="linear", fill_value="extrapolate")
            I_simulations["R"].append(list(f_R(t_bounds)))

            f_D = interp1d(results.t, results.y[3], kind="linear", fill_value="extrapolate")
            I_simulations["D"].append(list(f_D(t_bounds)))

#    I_simulations = np.array(I_simulations)
#    I_interval_l, I_interval_u = list(), list()
#    for k in range(len(t_bounds)):
#        I_interval_l.append(np.quantile(I_simulations[:,k], confidence_interval/2))
#        I_interval_u.append(np.quantile(I_simulations[:,k], 1-confidence_interval/2))
#    return t_bounds, I_interval_l, I_interval_u
        
    intervals_l = {"I": list(), "R": list(), "D": list()}
    intervals_u = {"I": list(), "R": list(), "D": list()}
    for I_type in ["I", "R", "D"]:

        I_simulations[I_type] = np.array(I_simulations[I_type])

        for k in range(len(t_bounds)):
            intervals_l[I_type].append(np.quantile(I_simulations[I_type][:,k], confidence_interval/2))
            intervals_u[I_type].append(np.quantile(I_simulations[I_type][:,k], 1-confidence_interval/2))

    return t_bounds, intervals_l, intervals_u

def get_bounds_SIAS(data_df, simulations, simulation_key, p_IA0, sir_parent_key=None, sias_parent_key=None, policy_date_i=None, policy_date_f=None, t_steps_ahead=150):

    from scipy.interpolate import interp1d
    
    I_simulations = {"I_S": list(), "I_A": list(), "R_A": list(), "R_S": list(), "D": list(), "I_A+R_A": list(), "I_S+R_S+D": list()}

    # Bounds
    params_bounds = dict()
    for param in ["beta", "gamma", "rho"]:
        params_bounds[param] = {"lower": np.quantile(simulations[sir_parent_key]["params_list"][param], confidence_interval/2),
                                "upper": np.quantile(simulations[sir_parent_key]["params_list"][param], 1-confidence_interval/2)}

    for param in ["beta_A", "gamma_A", "p"]:
        params_bounds[param] = {"lower": np.quantile(simulations[sias_parent_key]["params_list"][param], confidence_interval/2),
                                "upper": np.quantile(simulations[sias_parent_key]["params_list"][param], 1-confidence_interval/2)}

    # Valid simulation parameters (that all parameters are inside the bounds)
    valid_sir_params = list()
    num_iterations = len(simulations[sir_parent_key]["params_list"]["beta"])
    for k in range(num_iterations):
        if all([((simulations[sir_parent_key]["params_list"][param][k] > params_bounds[param]["lower"]) and (simulations[sir_parent_key]["params_list"][param][k] < params_bounds[param]["upper"])) for param in ["beta", "gamma", "rho"]]):
            valid_sir_params.append((simulations[sir_parent_key]["params_list"]["beta"][k],
                                     simulations[sir_parent_key]["params_list"]["gamma"][k],
                                     simulations[sir_parent_key]["params_list"]["rho"][k]))

    valid_sias_params = list()
    num_iterations = len(simulations[sias_parent_key]["params_list"]["beta_A"])
    for k in range(num_iterations):
        if all([((simulations[sias_parent_key]["params_list"][param][k] > params_bounds[param]["lower"]) and (simulations[sias_parent_key]["params_list"][param][k] < params_bounds[param]["upper"])) for param in ["beta_A", "gamma_A", "p"]]):
            valid_sias_params.append((simulations[sias_parent_key]["params_list"]["beta_A"][k],
                                      simulations[sias_parent_key]["params_list"]["gamma_A"][k],
                                      simulations[sias_parent_key]["params_list"]["p"][k]))

    #assert np.isclose(np.mean(simulations[sir_parent_key]["params_list"]["beta"]), simulations[simulation_key]["params"]["beta_S"])
    #assert np.isclose(np.mean(simulations[sir_parent_key]["params_list"]["gamma"]), simulations[simulation_key]["params"]["gamma_S"])
    #assert np.isclose(np.mean(simulations[sir_parent_key]["params_list"]["rho"]), simulations[simulation_key]["params"]["rho"])

    #assert np.isclose(np.mean(simulations[sias_parent_key]["params_list"]["beta_A"]), simulations[simulation_key]["params"]["beta_A"])
    #assert np.isclose(np.mean(simulations[sias_parent_key]["params_list"]["gamma_A"]), simulations[simulation_key]["params"]["gamma_A"])
    #assert np.isclose(np.mean(simulations[sias_parent_key]["params_list"]["p"]), simulations[simulation_key]["params"]["p"])

    t_bounds = None
    
    for beta_S_i, gamma_S_i, rho_i in valid_sir_params:
        for beta_A_i, gamma_A_i, p_i in valid_sias_params:

            if ("psi" in simulations[simulation_key]["params"]) and (not simulations[simulation_key]["params"]["psi"] is None) and (not np.isnan(simulations[simulation_key]["params"]["psi"])):
                psi_i = simulations[simulation_key]["params"]["psi"]
                date_i = policy_date_i if (not policy_date_i is None) else date_of_effectiveness_of_policy
                transmition_factor_fn = get_transmition_factor_fn(psi=psi_i, date_i=date_i, date_f=policy_date_f)
            else:
                transmition_factor_fn = None

            results = solve_ivp(fun=partial(diff_SIAS_with_p, beta_A=beta_A_i, beta_S=beta_S_i, gamma_A=gamma_A_i, gamma_S=gamma_S_i, mu=0, Lambda=0, rho=rho_i, p=p_i, N=data_df.iloc[1]["N"], transmition_factor_fn=transmition_factor_fn), t_span=(1, data_df["t"].max()+t_steps_ahead), y0=[S0-p_IA0, p_IA0, I0, 0, 0, 1], method="RK45", max_step=0.2)

            if (t_bounds is None):
                t_bounds = results.t

            if (len(results.t) == len(t_bounds)):
                I_simulations["I_A"].append(list(results.y[1]))
                I_simulations["I_S"].append(list(results.y[2]))
                I_simulations["R_A"].append(list(results.y[3]))
                I_simulations["R_S"].append(list(results.y[4]))
                I_simulations["D"].append(list(results.y[5]))
            else:
                f_A = interp1d(results.t, results.y[1], kind="linear", fill_value="extrapolate")
                I_simulations["I_A"].append(list(f_A(t_bounds)))
                
                f_S = interp1d(results.t, results.y[2], kind="linear", fill_value="extrapolate")
                I_simulations["I_S"].append(list(f_S(t_bounds)))

                f_RA = interp1d(results.t, results.y[3], kind="linear", fill_value="extrapolate")
                I_simulations["R_A"].append(list(f_RA(t_bounds)))

                f_RS = interp1d(results.t, results.y[4], kind="linear", fill_value="extrapolate")
                I_simulations["R_S"].append(list(f_RS(t_bounds)))

                f_D = interp1d(results.t, results.y[5], kind="linear", fill_value="extrapolate")
                I_simulations["D"].append(list(f_D(t_bounds)))

            I_simulations["I_A+R_A"].append(list(np.array(I_simulations["I_A"][-1])+np.array(I_simulations["R_A"][-1])))
            I_simulations["I_S+R_S+D"].append(list(np.array(I_simulations["I_S"][-1])+np.array(I_simulations["R_S"][-1])+np.array(I_simulations["D"][-1])))

    intervals_l = {"I_A": list(), "I_S": list(), "R_A": list(), "R_S": list(), "D": list(), "I_A+R_A": list(), "I_S+R_S+D": list()}
    intervals_u = {"I_A": list(), "I_S": list(), "R_A": list(), "R_S": list(), "D": list(), "I_A+R_A": list(), "I_S+R_S+D": list()}
    for I_type in ["I_A", "I_S", "R_A", "R_S", "D", "I_A+R_A", "I_S+R_S+D"]:

        I_simulations[I_type] = np.array(I_simulations[I_type])

        for k in range(len(t_bounds)):
            intervals_l[I_type].append(np.quantile(I_simulations[I_type][:,k], confidence_interval/2))
            intervals_u[I_type].append(np.quantile(I_simulations[I_type][:,k], 1-confidence_interval/2))
    
    return t_bounds, intervals_l, intervals_u


# # Data
# 
# We load the coronavirus CSV. Currently, we download it from:
# https://coronavirus.saude.gov.br/
# and then save in the data directory.
# 
# Last file saved is "COVID19_20200331.csv"

# In[7]:


data_file = sorted([file_name for file_name in os.listdir(data_dir) if (file_name[0:12].upper()=="COVID19_2020") and (file_name[-4:].lower()==".csv")])[-1]
coronavirus_df = pd.read_csv(path_join(data_dir, data_file), sep=";", converters={"data": lambda x: datetime.strptime(x, "%d/%m/%Y").date()})
coronavirus_df.rename(columns={"regiao": "region", "estado": "uf", "data": "date", "casosNovos": "cases_new", "casosAcumulados": "cases", "obitosNovos": "deaths_new", "obitosAcumulado": "deaths"}, inplace=True)

if (coronavirus_df["date"].max() < current_date):
    current_date = coronavirus_df["date"].max()
    #os.unlink(data_file)

coronavirus_df = pd.merge(left=coronavirus_df, right=brazil_states_df[["uf", "population"]], how="inner", on="uf")

if (coronavirus_df["date"].max() > current_date):
    coronavirus_df = coronavirus_df.loc[coronavirus_df["date"] <= current_date]
coronavirus_df["country"] = "BR"

coronavirus_df.fillna(0, inplace=True)

coronavirus_df.sort_values(["uf", "date"], inplace=True)
coronavirus_df["cases_diff"] = coronavirus_df.groupby(["uf"])["cases"].diff()

coronavirus_df.sort_values(["date", "uf"], inplace=True)


# In[8]:


N = brazil_states_df.loc[brazil_states_df["uf"]!="BR"]["population"].sum()

def get_coronavirus_data(state="BR", return_population = False):
    data_df = coronavirus_df.copy()

    if (state is None) or ((not state is None) and (state=="BR")):
        data_df = data_df.groupby("date")[["cases_new", "cases", "deaths_new", "deaths"]].sum()
    else:
        data_df = data_df.loc[(data_df["uf"]==state)].groupby("date")[["cases_new", "cases", "deaths_new", "deaths"]].sum()

    data_df = data_df.loc[data_df["cases"]>0].reset_index()
    data_df["t"] = data_df["date"].apply(lambda d: (d-data_df["date"].min()).days)
    data_df["recovers"] = data_df["cases"].diff().shift(14).cumsum()
    data_df.fillna(0, inplace=True)

    data_df["cases_diff"] = data_df["cases"].diff()

    _N = brazil_states_df.loc[brazil_states_df["uf"]==state].iloc[0]["population"]

    data_df["S"] = _N - data_df["cases"] - data_df["recovers"] - data_df["deaths"]
    data_df["I"] = data_df["cases"] - data_df["recovers"] # stock of infected
    data_df["R"] = data_df["recovers"] # stock of recovered population
    data_df.fillna(0, inplace=True)
    data_df["N"] = data_df["S"] + data_df["I"] + data_df["R"]

    data_df["D"] = data_df["deaths"]

    if (return_population):
        return data_df, _N

    return data_df

def t2date_in_df(t,pos=None,df=None):
    first_date = df.iloc[0]["date"]
    return first_date if np.isclose(t, 0) else first_date + timedelta(days=int(t))

variable = "cases"
all_variables = [variable, "deaths", "cases_diff"]

for state in list(coronavirus_df["uf"].unique()) + ["BR"]:
    if (state != state_being_analysed):
        continue

    try:
        subset_df = get_coronavirus_data(state=state)
        subset_df = subset_df.loc[subset_df[variable] > 0]
        
        print(subset_df)
        #display(HTML(tabulate(subset_df.reset_index(), headers=subset_df.reset_index().columns, tablefmt='html')))

        plt.figure()
        ax = subset_df[variable].plot(title=variable, logy=False)
        ax.set_xlabel("Date")
        ax.set_ylabel(variable)
        
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(partial(t2date_in_df, df=subset_df)))
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)

        plt.show()

    except:
        pass


# # Results

# In[9]:


data_df = get_coronavirus_data(state=state_being_analysed)
print(data_df[["date", "cases", "cases_diff"]])

S0, I0 = data_df.iloc[0][["S", "I"]]

def plot_ID(df, t, I, R, D, x_limits=None):

    if (x_limits is None):
        x_limits = (1, df["t"].max())

    if (D is None):
        fig, ax_I = plt.subplots(nrows=1, ncols=1)
        ax_D = None
        ax = ax_I
    else:
        fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True)
        ax_I, ax_D = axs
        ax = ax_D

    # Infected    
    line_I_real, = ax_I.plot(df["t"], df["cases"], label="Real", color="DarkOrange", marker="*", linestyle="")
    I_max = df.loc[(df["t"] >= x_limits[0]) & (df["t"] <= x_limits[1])]["I"].max()

    ax_I.plot(t, I+R+D, linestyle="-", color="DodgerBlue")
    
    ax_I.set_ylim(0, 1.1*I_max)
    ax_I.set_ylabel("Infected")
    ax_I.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))

    # Deads
    if (not D is None):
        ax_D.plot(df["t"], df["D"], label="Real", color="DarkOrange", marker="*", linestyle="")
        D_max = df.loc[(df["t"] >= x_limits[0]) & (df["t"] <= x_limits[1])]["D"].max()
    
        ax_D.plot(t, D, linestyle="-", color="DodgerBlue")
        
        ax_D.set_ylim(0, 1.1*D_max)
        ax_D.set_ylabel("Deaths")
        ax_D.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))

    #ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
    ax.set_xlim(*x_limits)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    plt.show()

    return


# ## Part 1: SID model, eq. (3)
# 
# Calculate parameters for SID model:
# 
# \begin{equation}
#     \begin{array}{rcl}
#     \displaystyle \frac{dS}{dt} & = &  - \frac{\beta I S}{N}\\[3mm]
#     \displaystyle \frac{dI}{dt} & = & \frac{\beta I S}{N} - \gamma I  - \frac{\rho}{1 - \rho} \gamma I = \frac{\beta I S}{N} - \frac{\gamma I}{1 - \rho}\\[3mm]
#     \displaystyle \frac{dD}{dt}&=&\frac{\rho}{1-\rho} \gamma I\\
#    % \frac{dR}{dt} & = &  \gamma I\\
#     \end{array}
# \label{eq:SIRDead}
# \end{equation}
# 
# Data from February 26, 2020, to March 22, 2020.

# In[11]:

simulations = dict()

num_iterations = 100

if (os.path.exists(simulation_csv)):
    sir_params_df = pd.read_csv(simulation_csv, sep=";", converters={"date": lambda x: date(*tuple(map(int, x.split("-"))))})
else:
    sir_params_df = pd.DataFrame(columns = ["description", "cut_date", "beta", "gamma", "rho", "psi", "R0", "beta_list", "gamma_list", "rho_list", "psi_list"])

list_params = ["beta", "gamma", "rho", "psi"]
for list_param in list_params:
    if (not list_param+"_list" in sir_params_df.columns):
        sir_params_df[list_param+"_list"] = None
    else:
        sir_params_df[list_param+"_list"] = sir_params_df[list_param+"_list"].apply(lambda x: list(map(float, x.strip("[]").split(", "))) if isinstance(x, str) and (len(x)>0) and (x!="[]") else None)

if (sir_params_df.loc[sir_params_df["description"].str.contains("SIR with social distance") & np.isnan(sir_params_df["psi"])].shape[0] > 0):
    sir_params_df.loc[sir_params_df["description"].str.contains("SIR with social distance") & np.isnan(sir_params_df["psi"]), "psi_list"] = [1]
    sir_params_df.loc[sir_params_df["description"].str.contains("SIR with social distance") & np.isnan(sir_params_df["psi"]), "psi"] = 1

for time_delta in [0, 4, (date(2020, 3, 30) - date_of_effectiveness_of_policy).days]: #np.arange(start=0, stop=9):

    np.random.seed(random_seed)

    cut_date = date_of_effectiveness_of_policy
    if (time_delta > 0):
        cut_date += timedelta(days=int(time_delta))
    
    print(cut_date)

    description = get_SIR_description(cut_date)
    simulations[description] = {"params": {"beta": None, "gamma": None, "Lambda": 0, "mu": 0, "rho": None}, "cut_date": cut_date, "params_list": dict()}
    
    df = data_df.copy()
    df = df.loc[df["date"] <= cut_date]

    simulation_row = sir_params_df.loc[sir_params_df["description"] == description].iloc[0] if (sir_params_df.loc[sir_params_df["description"] == description].shape[0] > 0) else None
    if (simulation_row is None):
        params = simulations[description]["params"]

        x0 = None
        variables_to_optimize = ["I", "D"]
        optimization_params, params_list = get_SIR_parameters_from_optimization(df, beta=params["beta"], gamma=params["gamma"], mu=params["mu"], Lambda=params["Lambda"], rho=params["rho"], x0 = x0, variables_to_optimize = variables_to_optimize, num_iterations = num_iterations, return_params_list = True, plot_params_histogram = False)
        for param_name, param_value in params.items():
            if (param_value is None) and (param_name in optimization_params):
                simulations[description]["params"][param_name] = optimization_params[param_name]
                #simulations[description]["params"][param_name] = np.quantile(params_list[param_name], 0.5)
                simulations[description]["params_list"][param_name] = params_list[param_name]

        R0 = simulations[description]["params"]["beta"] * (1 - simulations[description]["params"]["rho"]) / simulations[description]["params"]["gamma"]
        sir_params_df.loc[len(sir_params_df)] = [description, cut_date, simulations[description]["params"]["beta"], simulations[description]["params"]["gamma"], simulations[description]["params"]["rho"], None, R0, simulations[description]["params_list"]["beta"], simulations[description]["params_list"]["gamma"], simulations[description]["params_list"]["rho"], None]
        sir_params_df.to_csv(simulation_csv, sep=";", header = True, mode = 'w', index=False)
    else:
        for param_name, param_value in simulations[description]["params"].items():
            if (param_name+"_list" in sir_params_df.columns) and (not simulation_row[param_name+"_list"] is None) and isinstance(simulation_row[param_name+"_list"], list):
                simulations[description]["params_list"][param_name] = simulation_row[param_name+"_list"]
            if (param_value is None):
                simulations[description]["params"][param_name] = simulation_row[param_name]
                #simulations[description]["params"][param_name] = np.quantile(simulation_row[param_name+"_list"], 0.5)

    #simulations[description]["results"] = solve_ivp(fun=partial(diff_SIR, **simulations[description]["params"], N=df.iloc[0]["N"], t_train=df["t"].tolist(), N_train=df["N"].tolist()), t_span=(1, data_df["t"].max()+150), y0=[S0, I0, 0, 0], method="RK45", max_step=0.1)
    simulations[description]["results"] = simulate_SIR(simulations, key=description, df=data_df, use_median = use_median_for_parameters, n_steps_ahead = 150)

    plot_ID(df, t=simulations[description]["results"].t, I=simulations[description]["results"].y[1], R=simulations[description]["results"].y[2], D=simulations[description]["results"].y[-1])

    print(description + " parameters: ", simulations[description]["params"], "\n")


# Differences in trajectories of the SIR model without social distance.
# 
# **Figure 1**. Estimations of the SIR model for different final date points. The solid line corresponds to the last date which the model was estimated, and the dashed line are model predictions. We represent the real data as points.

# In[ ]:


if (generate_figures):
    num_steps_ahead = 1
    last_date, last_t = data_df["date"].max(), data_df["t"].max()
    print("last_date = ", last_date)
    fig_num = 1
    for x_limits in [(data_df["t"].min(), last_t+num_steps_ahead), (22, last_t)]:
        fig, ax = plt.subplots(nrows=1, ncols=1)
    
        line_I_real, = ax.plot(data_df["t"], data_df["cases"], label="Real", color="DarkOrange", marker="*", linestyle="")
        I_max = data_df["cases"].max()
    
        colors = ["LightSteelBlue", "DodgerBlue", "Navy"]
        #for k, cut_date in enumerate([date_of_effectiveness_of_policy, date_of_effectiveness_of_policy+timedelta(days=4), date(2020, 3, 30)]):
        for k, cut_date in enumerate([date_of_effectiveness_of_policy, date(2020, 3, 30)]):
            description = get_SIR_description(cut_date)
    
            t, I, R, D = simulations[description]["results"].t, simulations[description]["results"].y[1], simulations[description]["results"].y[2], simulations[description]["results"].y[-1]
    
            t_cut = date2t(cut_date)
            t_training_indexes = [i for i, ti in enumerate(t) if (ti <= t_cut)]
            ax.plot(t[t_training_indexes], list(np.array(I[t_training_indexes])+np.array(R[t_training_indexes])+np.array(D[t_training_indexes])), label="Last data: " + cut_date.strftime("%d %b, %Y"), linestyle="-", color=colors[k])
    
            t_prediction_indexes = [i for i, ti in enumerate(t) if (ti > t_cut) and (ti <= last_t+num_steps_ahead)]
            ax.plot(t[t_prediction_indexes], list(np.array(I[t_prediction_indexes])+np.array(R[t_prediction_indexes])+np.array(D[t_prediction_indexes])), linestyle="--", color=colors[k])
    
            if (max(I[t_prediction_indexes])+max(R[t_prediction_indexes]) > I_max):
                I_max = max(I[t_prediction_indexes])+max(R[t_prediction_indexes])
    
        ax.set_xlim(*x_limits)
        ax.set_ylim(0, min(I_max, 8000))
        ax.set_ylabel("Cumulative number of infected cases")
    
        ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
        plt.savefig(output_dir+"fig_SIR_unrestricted_predictions"+str(fig_num)+".png", dpi=300, bbox_inches='tight')
        plt.show()
        
        fig_num += 1


# **Table 1**. Estimated values of the epidemiological parameters.
# 
# We show parameters $\beta$, $\gamma$ and $\rho$ of the SIR model, before the "structural break".

# In[ ]:

sir_model_key = get_SIR_description(cut_date=date_of_effectiveness_of_policy)

chosen_models = collections.OrderedDict()
cut_date = date_of_effectiveness_of_policy
description = sir_model_key

sir_params_especification_df = pd.DataFrame(columns = ["param", "q", "value"])

params_sir = dict()
params_sir["mean"] = {"beta": simulations[description]["params"]["beta"],
                      "beta_std": np.std(simulations[description]["params_list"]["beta"]),
                      "gamma": simulations[description]["params"]["gamma"],
                      "gamma_std": np.std(simulations[description]["params_list"]["gamma"]),
                      "rho": simulations[description]["params"]["rho"],
                      "rho_std": np.std(simulations[sir_model_key]["params_list"]["rho"])}

for param in ["beta", "gamma", "rho"]:
    sir_params_especification_df.loc[len(sir_params_especification_df)] = [param, "mean", params_sir["mean"][param]]

R0_mean = params_sir["mean"]["beta"]*(1-params_sir["mean"]["rho"])/(params_sir["mean"]["gamma"]+mu)
sir_params_especification_df.loc[len(sir_params_especification_df)] = ["R0", "mean", R0_mean]

for q in [0.025, 0.05, 0.5, 0.95, 0.975]:
    beta_list = simulations[description]["params_list"]["beta"]
    gamma_list = simulations[description]["params_list"]["gamma"]
    rho_list = simulations[description]["params_list"]["rho"]
    
    params_sir["quantile-"+str(q)] = {"beta": np.quantile(beta_list, q),
                                      "gamma": np.quantile(gamma_list, q),
                                      "rho": np.quantile(rho_list, q),
                                      "R0": np.quantile([beta*(1-rho)/(gamma+mu) for beta, gamma, rho in zip(beta_list, gamma_list, rho_list)], q)}

    for param, value in params_sir["quantile-"+str(q)].items():
        sir_params_especification_df.loc[len(sir_params_especification_df)] = [param, q, value]

chosen_models = collections.OrderedDict()
chosen_models[description] = "SIR without social distance (Last data: " + cut_date.strftime("%d %b, %Y") + ")"

sir_params_especification_df = sir_params_especification_df.set_index(["param", "q"]).unstack()

print(sir_params_especification_df)

sir_params_especification_df.to_csv(path_join(output_dir, "sir_params.csv"), sep=";", header = True, mode = 'w', index=True)


# **Appendix**. Random seed vs. Estimated values.
# 

# In[ ]:


for least_squares_f_scale in [args.least_squares_f_scale]:
    print(50*"*")
    print("f_scale = ", least_squares_f_scale)
    print(50*"-")

    params_rand_df = pd.DataFrame(columns=["random_seed", "beta", "gamma", "rho"])
    for name in [name for name in os.listdir(base_output_dir) if os.path.isdir(path_join(base_output_dir, name)) and (name[0:5]=="rand-")]:
        seed = int(name.split("_")[0][5:])
        if np.isclose(least_squares_f_scale, 1.0) and ("_fscale-" in name):
            continue
        elif (not np.isclose(least_squares_f_scale, 1.0)) and (not "rand-"+str(seed)+"_fscale-"+str(least_squares_f_scale) in name):
            continue
    
        simulation_dir = path_join(base_output_dir, "rand-"+str(seed)+dir_separator) if np.isclose(least_squares_f_scale, 1.0) else path_join(base_output_dir, "rand-"+str(seed)+"_fscale-"+str(least_squares_f_scale)+dir_separator)
        simulation_file = simulation_dir + "simulations-"+state_being_analysed+"_loss-"+args.least_squares_loss+".csv"
        if (not os.path.exists(simulation_file)):
            continue
        
        df = pd.read_csv(simulation_file, sep=";")
        row = df.iloc[0]
        params_rand_df.loc[len(params_rand_df)] = [seed, row["beta"], row["gamma"], row["rho"]]
    
    params_rand_df.sort_values("random_seed", inplace=True)
    params_rand_df["random_seed"] = params_rand_df["random_seed"].apply(lambda x: "{:d}".format(int(x)))
    print(tabulate(params_rand_df, headers='keys', tablefmt='psql', showindex=False))
    
    params_rand_df["gamm_dist"] = params_rand_df["gamma"].apply(lambda x: np.abs(x - np.quantile(params_rand_df["gamma"], 0.5)))

    print(params_rand_df.loc[params_rand_df["gamm_dist"]==params_rand_df["gamm_dist"].min()])

    print("")


# Calculate parameter $\psi$ for SIR model with government confinement measures.
# 
# \begin{equation}
#     \begin{array}{rcl}
#     \displaystyle \frac{dS}{dt} & = &  - \psi \frac{\beta I S}{N}\\[3mm]
#     \displaystyle \frac{dI}{dt} & = & \psi \frac{\beta I S}{N} - \gamma I  - \frac{\rho}{1 - \rho} \gamma I = \psi \frac{\beta I S}{N} - \frac{\gamma I}{1 - \rho}\\[3mm]
#     \qqdisplaystyle \frac{dD}{dt}&=& \frac{\rho}{1-\rho} \gamma I\\
#    % \frac{dR}{dt} & = &  \gamma I\\
#     \end{array}
# \label{eq:SIRDead}
# \end{equation}
# 
# Data from February 26, 2020, to March 30, 2020.

# In[ ]:

num_iterations = 100

for time_delta in np.arange(start=0, stop=(date(2020, 3, 30) - date_of_effectiveness_of_policy).days):

    np.random.seed(random_seed)

    cut_date = policy_begin_date
    if (time_delta > 0):
        cut_date += timedelta(days=int(time_delta))

    description = "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
    #simulations[description] = {"params": {"beta": simulations[sir_model_key]["params"]["beta"], "gamma": simulations[sir_model_key]["params"]["gamma"], "Lambda": simulations[sir_model_key]["params"]["Lambda"], "mu": simulations[sir_model_key]["params"]["mu"], "rho": simulations[sir_model_key]["params"]["rho"], "is_policy_effective_fn": lambda t: (t>date2t(date_of_effectiveness_of_policy))}, "params_list": dict()}
    simulations[description] = {"params": {"beta": simulations[sir_model_key]["params"]["beta"], "gamma": simulations[sir_model_key]["params"]["gamma"], "Lambda": simulations[sir_model_key]["params"]["Lambda"], "mu": simulations[sir_model_key]["params"]["mu"], "rho": simulations[sir_model_key]["params"]["rho"], "is_policy_effective_fn": lambda t: (t>date2t(date_of_effectiveness_of_policy))},
                                "params_list": {"beta": simulations[sir_model_key]["params_list"]["beta"], "gamma": simulations[sir_model_key]["params_list"]["gamma"], "rho": simulations[sir_model_key]["params_list"]["rho"]}}

    df = data_df.copy()
    df = df.loc[df["date"] <= cut_date]

    simulation_row = sir_params_df.loc[sir_params_df["description"] == description].iloc[0] if (sir_params_df.loc[sir_params_df["description"] == description].shape[0] > 0) else None
    if (simulation_row is None):

        params = simulations[description]["params"]

        optimization_params, params_list = get_SIR_parameters_from_optimization(df, beta=params["beta"], gamma=params["gamma"], mu=params["mu"], Lambda=params["Lambda"], rho=params["rho"], is_policy_effective_fn=params["is_policy_effective_fn"], variables_to_optimize = ["I", "D"], num_iterations = num_iterations, return_params_list = True, plot_params_histogram = False)
        #optimization_params, params_list = get_SIR_parameters_from_optimization(df, beta=params["beta"], gamma=params["gamma"], mu=params["mu"], Lambda=params["Lambda"], rho=None, is_policy_effective_fn=params["is_policy_effective_fn"], variables_to_optimize = ["I"], num_iterations = num_iterations, return_params_list = True, plot_params_histogram = True)
        simulations[description]["params"]["psi"] = optimization_params["psi"]
        simulations[description]["params_list"]["psi"] = params_list["psi"]

        #simulations[description]["params"]["rho"] = optimization_params["rho"]
        #simulations[description]["params_list"]["rho"] = params_list["rho"]

        R0 = simulations[description]["params"]["beta"]*simulations[description]["params"]["psi"] * (1 - simulations[description]["params"]["rho"]) / simulations[description]["params"]["gamma"]
        sir_params_df.loc[len(sir_params_df)] = [description, cut_date, simulations[description]["params"]["beta"], simulations[description]["params"]["gamma"], simulations[description]["params"]["rho"], simulations[description]["params"]["psi"], R0, simulations[sir_model_key]["params_list"]["beta"], simulations[sir_model_key]["params_list"]["gamma"], simulations[sir_model_key]["params_list"]["rho"], simulations[description]["params_list"]["psi"]]

        sir_params_df.to_csv(simulation_csv, sep=";", header = True, mode = 'w', index=False)
    else:
        for param_name, param_value in simulations[description]["params"].items():
            if (param_value is None):
                simulations[description]["params"][param_name] = simulation_row[param_name]
            if (param_name+"_list" in sir_params_df.columns) and (not simulation_row[param_name+"_list"] is None) and isinstance(simulation_row[param_name+"_list"], list):
                simulations[description]["params_list"][param_name] = simulation_row[param_name+"_list"]
        simulations[description]["params"]["psi"] = simulation_row["psi"]

    print(description + " parameters: ", simulations[description]["params"], "\n")

    #simulations[description]["params"]["transmition_factor_fn"] = get_transmition_factor_fn(psi=simulations[description]["params"]["psi"], date_i=date_of_effectiveness_of_policy)
    #params = {key: value for key,value in simulations[description]["params"].items() if (not key in ["psi", "is_policy_effective_fn"])}
    #simulations[description]["results"] = solve_ivp(fun=partial(diff_SIR, **params, N=df.iloc[1]["N"], t_span=(1, data_df["t"].max()+150), y0=[S0, I0, 0, 0], method="RK45", max_step=0.1)

    simulations[description]["results"] = simulate_SIR(simulations, key=description, df=data_df, use_median = use_median_for_parameters, n_steps_ahead = 150)

# **Figure 2**. Short term forecast of the SID model taking into account government social distance measures. The solid line corresponds to the last date which the model was estimated, and the dashed line are model predictions. We show the evolution of the infected with 95% confidence interval. We represent the real data as points.

# In[ ]:


if (generate_figures):
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)
    
    num_steps_ahead = 3
    
    t_bounds, intervals_l, intervals_u = get_bounds_SIR(data_df, simulations, description, parent_key=sir_model_key)
    
    color_mean = "Sienna"
    color_interval = "LightCoral"
    
    fig_num = 1
    num_steps_ahead = 1
    for x_limits in [(data_df["t"].min(), last_t+num_steps_ahead), (22, last_t+num_steps_ahead)]:
        fig, ax = plt.subplots(nrows=1, ncols=1)
    
        line_I_real, = ax.plot(data_df["t"], data_df["cases"], label="Real", color="DarkOrange", marker="*", linestyle="")
        I_max = data_df["I"].max()
    
        colors = ["RosyBrown", "SaddleBrown"]
        descriptions = []
        for k, cut_date in enumerate([policy_begin_date, date(2020, 3, 30)]):
            description = "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
            
            results = simulate_SIR(simulations, key=description, df=data_df, use_median = False, n_steps_ahead = 2)
            t, I, R, D = results.t, results.y[1], results.y[2], results.y[3]
    
            if (cut_date == date(2020, 3, 30)):
                ax.fill_between(t_bounds, np.array(intervals_u["I"])+np.array(intervals_u["R"])+np.array(intervals_u["D"]), np.array(intervals_l["I"])+np.array(intervals_l["R"])+np.array(intervals_l["D"]), color=color_interval, alpha=0.2)
    
            t_cut = date2t(cut_date)
            t_training_indexes = [i for i, ti in enumerate(t) if (ti <= t_cut)]
            t_prediction_indexes = [i for i, ti in enumerate(t) if (ti > t_cut) and (ti <= last_t+num_steps_ahead)]
    
            ax.plot(t[t_training_indexes], I[t_training_indexes]+R[t_training_indexes]+D[t_training_indexes], label="Last data: " + cut_date.strftime("%d %b, %Y"), linestyle="-", color=colors[k])
    
            t_prediction_indexes = [i for i, ti in enumerate(t) if (ti > t_cut) and (ti <= last_t+num_steps_ahead)]
            ax.plot(t[t_prediction_indexes], I[t_prediction_indexes]+R[t_prediction_indexes]+D[t_prediction_indexes], linestyle="--", color=colors[k])
    
            if (max(I[t_prediction_indexes]) > I_max):
                I_max = max(I[t_prediction_indexes])
    
        ax.set_xlim(*x_limits)
        ax.set_ylim(0, min(I_max, 8000))
        ax.set_ylabel("Cumulative number of infected cases")
    
        ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    
        plt.savefig(output_dir+"fig_SIR_restricted_predictions"+str(fig_num)+".png", dpi=300, bbox_inches='tight')
        plt.show()
        
        fig_num += 1

psi_df = pd.DataFrame(columns = ["t", "date", "psi", "R"])
for k in np.arange(start=1, stop=(date(2020, 3, 30) - date_of_effectiveness_of_policy).days+1):
    cut_date = date_of_effectiveness_of_policy + timedelta(days=int(k))
    description = "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
    rho_k = simulations[description]["params"]["psi"]*simulations[description]["params"]["beta"]*(1-simulations[description]["params"]["rho"])/simulations[description]["params"]["gamma"]
    psi_df.loc[len(psi_df)] = [date2t(cut_date), cut_date, simulations[description]["params"]["psi"], rho_k]
#display(HTML(tabulate(psi_df, headers=psi_df.columns, tablefmt='html')))
print(tabulate(psi_df, headers='keys', tablefmt='psql', showindex=False))

if (generate_figures):
    from matplotlib import rc
    rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    rc('text', usetex=True)
    
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(psi_df["t"], psi_df["psi"], linestyle="-", marker="o")
    ax.set_xticks(psi_df["t"].values)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    ax.set_ylabel(r"$\mathbf{\psi}$", fontsize=20)
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    plt.savefig(output_dir+"fig_SIR_restricted_psi.png", dpi=300, bbox_inches='tight')
    plt.show()
    
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)

cut_date = date(2020, 3, 30)
description = "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
chosen_models[description] = "SIR with social distance (Policy effect: " + policy_begin_date.strftime("%d %b, %Y") + ", last data: " + cut_date.strftime("%d %b, %Y") + ")"


# Generating other scenarios:
# - (I) no measures of confinement policy (black line);
# - (II) current confinement policy imposed by the government for an indefinite time (blue line);
# - (III) 2-month confinement policy imposed by the government (yellow line); and
# - (IV) optimum limited time confinement policy imposed by the government, so that the second infection peak is not greater than cenario II (red line)
# 
# **Figure 4**. Long term forecasts of number of infected for different scenarios using the SID model. Black, blue, yellow and red lines represent scenarios I to IV, respectively.

# In[ ]:


if (generate_figures):
    cut_date = date(2020, 3, 30)
    sir_model_with_sd = "SIR with social distance (policy start: " + policy_begin_date.isoformat() +  ", cut: " + cut_date.isoformat() + ")"
    
    for description, policy_end_date in [("SIR with social distance for 2 months", policy_begin_date+timedelta(days=60)), ("SIR with social distance, optimized peak", date(2020, 6, 22))]:
        print("\n")
        print("policy_end_date = ", policy_end_date)
    
        chosen_models[description] = description
    
        simulations[description] = {"params": {k: v for k, v in simulations[sir_model_with_sd]["params"].items()}}
        simulations[description]["params"]["transmition_factor_fn"] = get_transmition_factor_fn(psi=simulations[sir_model_with_sd]["params"]["psi"], date_i=date_of_effectiveness_of_policy, date_f=policy_end_date)
        simulations[description]["params"]["is_policy_effective_fn"] = get_is_policy_effective_fn(date_i=date_of_effectiveness_of_policy, date_f=policy_end_date)
        simulations[description]["results"] = solve_ivp(fun=partial(diff_SIR, **{k: v for k, v in simulations[description]["params"].items() if (not k in ["psi", "is_policy_effective_fn"])}, N=data_df.iloc[0]["N"]), t_span=(1, data_df["t"].max()+150), y0=[S0, I0, 0, 1], method="RK45", max_step=0.1)

    import locale
    locale.setlocale(locale.LC_ALL, '')  # Use '' for auto, or force e.g. to 'en_US.UTF-8'
    
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True)
    colors = ["Black", "Blue", "Orange", "Red"]
    cenarios_df = pd.DataFrame(columns=["description", "peak at", "%peak"])
    
    sir_model_with_sd = get_SIR_description(cut_date=date(2020, 3, 30), policy_begin_date=policy_begin_date)
    
    for i, (model_key_to_extract_params, label, policy_date_i, policy_date_f) in enumerate([(sir_model_key, "I", None, None),
                                                                                            (sir_model_with_sd, "II", date_of_effectiveness_of_policy, None),
                                                                                            (sir_model_with_sd, "III", date_of_effectiveness_of_policy, date_of_effectiveness_of_policy+timedelta(days=60)),
                                                                                            (sir_model_with_sd, "IV", date_of_effectiveness_of_policy, date(2020, 6, 22))]):
        results = simulate_SIR(simulations, key=model_key_to_extract_params, df=data_df, use_median=True, n_steps_ahead=150, policy_date_i=policy_date_i, policy_date_f=policy_date_f)
        t, I, D = results.t, results.y[1], results.y[-1]
    
        SIR_statistics = get_SIR_statistics(t, I)
        cenarios_df.loc[len(cenarios_df)] = [label + " (" + colors[i] + ")", SIR_statistics["peak_date"], 100*SIR_statistics["peak_percent"]]
    
        t_bounds, intervals_l, intervals_u = get_bounds_SIR(data_df, simulations, simulation_key=model_key_to_extract_params, parent_key=sir_model_key, policy_date_i=policy_date_i, policy_date_f=policy_date_f)
        ax.fill_between(t_bounds, intervals_u["I"], intervals_l["I"], color=colors[i], alpha=0.2)
    
        line_I, = ax.plot(t, I, label=label, color=colors[i])
        ax.set_ylabel("Infected")
    
    ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1, loc='upper right')
    
    ax.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))
    
    ax.set_xlim(40, 140)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    
    plt.savefig(output_dir+"fig_SIR_models_comparison.png", dpi=300, bbox_inches='tight')
    
    plt.show()
    
    print(tabulate(cenarios_df, headers='keys', tablefmt='psql', showindex=False))
    #display(HTML(tabulate(cenarios_df, headers=cenarios_df.columns, tablefmt='html')))


# ## Part 2: SIASD model, eq. (4)
# 
# We calculate the parameters of the SIASD model.
# 
# \begin{equation}
# \begin{array}{rcl}
#     \displaystyle \frac{dS}{dt} & = & \displaystyle  - (\beta_A I_A + \beta_S I_S)\frac{ S}{N} \\[3mm]
#     \displaystyle \frac{dI_A}{dt} & = & \displaystyle  (1-p)(\beta_A I_A + \beta_S I_S)\frac{ S}{N} - (\gamma_A) I_A \\[3mm]
#     \displaystyle \frac{dI_S}{dt} & = & \displaystyle p (\beta_A I_A + \beta_S I_S)\frac{ S}{N} - \gamma_S I_S - \frac{\rho}{1 - \rho} \gamma_S I_S = p (\beta_A I_A + \beta_S I_S)\frac{ S}{N} - \frac{\gamma_S I_S}{1 - \rho} \\[3mm]
#     \displaystyle \frac{dD}{dt} & = & \displaystyle \frac{\rho}{1 - \rho} \gamma_S I_S \\[3mm]
# \end{array}
# \label{eq:IAISwithP} 
# \end{equation}
# 

# In[ ]:

rho = None
num_iterations = 100
simulation_sias_file = simulation_csv.replace("-"+state_being_analysed+"_loss-"+args.least_squares_loss+".csv", "_sias-"+state_being_analysed+"_loss-"+args.least_squares_loss+".csv")

if (os.path.isfile(simulation_sias_file)):
    sias_params_df = pd.read_csv(simulation_sias_file, sep=";", converters={"date": lambda x: date(*tuple(map(int, x.split("-"))))})
else:
    sias_params_df = pd.DataFrame(columns = ["description", "cut_date", "p_IA0", "beta_A", "beta_S", "gamma_A", "gamma_S", "p", "rho", "psi", "R0", "beta_A_list", "beta_S_list", "gamma_A_list", "gamma_S_list", "p_list", "rho_list", "psi_list"])

if (sias_params_df.loc[sias_params_df["description"].str.contains("SIR with social distance") & np.isnan(sias_params_df["psi"])].shape[0] > 0):
    sias_params_df.loc[sias_params_df["description"].str.contains("SIR with social distance") & np.isnan(sias_params_df["psi"]), "psi_list"] = [1]
    sias_params_df.loc[sias_params_df["description"].str.contains("SIR with social distance") & np.isnan(sias_params_df["psi"]), "psi"] = 1

list_params = ["beta_A", "beta_S", "gamma_A", "gamma_S", "p", "rho", "psi"]
for list_param in list_params:
    if (not list_param+"_list" in sias_params_df.columns):
        sias_params_df[list_param+"_list"] = None
    else:
        sias_params_df[list_param+"_list"] = sias_params_df[list_param+"_list"].apply(lambda x: list(map(float, x.strip("[]").split(", "))) if isinstance(x, str) else None)

p_IA0_list = [0.05, 0.10, 0.20, 0.25, 0.30, 0.50, 1.0, 1.5, 2.0, 3.0, 4.0]
for p_IA0 in p_IA0_list:
    for consider_policy_effect in [False, True]:
        if (consider_policy_effect):
            is_policy_effective_fn = lambda t: (t>date2t(date_of_effectiveness_of_policy))
        else:
            is_policy_effective_fn = None

        for time_delta in np.arange(start=0, stop=(date(2020, 3, 30) - date_of_effectiveness_of_policy).days+1):

            np.random.seed(random_seed)

            cut_date = date_of_effectiveness_of_policy
            if (time_delta > 0):
                cut_date += timedelta(days=int(time_delta))

            if (consider_policy_effect and (cut_date < policy_begin_date)):
                continue

            if (not consider_policy_effect) and (not cut_date in [date_of_effectiveness_of_policy]):
                continue

            if (consider_policy_effect):
                source_model = get_AS_description(cut_date=date_of_effectiveness_of_policy, p_IA0=p_IA0, policy_begin_date=None, rho=rho)
                description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=rho)
                simulations[description] = {"params": {"beta_A": simulations[source_model]["params"]["beta_A"], "beta_S": simulations[source_model]["params"]["beta_S"], "gamma_A": simulations[source_model]["params"]["gamma_A"], "gamma_S": simulations[source_model]["params"]["gamma_S"], "Lambda": simulations[source_model]["params"]["Lambda"], "mu": simulations[source_model]["params"]["mu"], "rho": simulations[source_model]["params"]["rho"], "p": simulations[source_model]["params"]["p"]}, "cut_date": cut_date, "params_list": dict()}
            else:
                description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=None, rho=rho)
                simulations[description] = {"params": {"beta_A": None, "beta_S": simulations[sir_model_key]["params"]["beta"], "gamma_A": None, "gamma_S": simulations[sir_model_key]["params"]["gamma"], "Lambda": 0, "mu": 0, "rho": simulations[sir_model_key]["params"]["rho"], "p": None}, "cut_date": cut_date, "params_list": dict()}

            df = data_df.copy()
            df = df.loc[df["date"] <= cut_date]

            simulation_row = sias_params_df.loc[sias_params_df["description"] == description].iloc[0] if (sias_params_df.loc[sias_params_df["description"] == description].shape[0] > 0) else None
            if (simulation_row is None):
                params = simulations[description]["params"]

                if (consider_policy_effect):
                    variables_to_optimize = ["I"]
                else:
                    variables_to_optimize = ["I", "D"]
                x0 = None
                optimization_params, params_list = get_SIAS_with_p_parameters_from_optimization(data_df=df, beta_A=simulations[description]["params"]["beta_A"], beta_S=simulations[description]["params"]["beta_S"], gamma_A=simulations[description]["params"]["gamma_A"], gamma_S=simulations[description]["params"]["gamma_S"], mu=simulations[description]["params"]["mu"], Lambda=simulations[description]["params"]["Lambda"], rho=simulations[description]["params"]["rho"], p=simulations[description]["params"]["p"], p_IA0=p_IA0, is_policy_effective_fn = is_policy_effective_fn, x0 = x0, force_same_beta = False, variables_to_optimize = variables_to_optimize, num_iterations = num_iterations, return_params_list = True, plot_params_histogram = False)
                for name in simulations[description]["params"].keys():
                    if (simulations[description]["params"][name] is None) and (name in optimization_params):
                        simulations[description]["params"][name] = optimization_params[name]
                        simulations[description]["params_list"][name] = params_list[name]

                simulations[description]["params"]["psi"] = optimization_params["psi"] if ("psi" in optimization_params) else None
                simulations[description]["params_list"]["psi"] = params_list["psi"] if ("psi" in params_list) else None

                RA = (1-simulations[description]["params"]["p"]) * simulations[description]["params"]["beta_A"] / (simulations[description]["params"]["gamma_A"] + simulations[description]["params"]["mu"])
                RS = simulations[description]["params"]["p"] * simulations[description]["params"]["beta_S"] / (simulations[description]["params"]["gamma_S"] + simulations[description]["params"]["mu"])
                R0 = RA + RS

                list_params_in_simulation = [simulations[description]["params_list"].get(p, None) for p in ["beta_A", "beta_S", "gamma_A", "gamma_S", "p", "rho", "psi"]]
                sias_params_df.loc[len(sias_params_df)] = [description, cut_date, p_IA0, simulations[description]["params"]["beta_A"], simulations[description]["params"]["beta_S"], simulations[description]["params"]["gamma_A"], simulations[description]["params"]["gamma_S"], simulations[description]["params"]["p"], simulations[description]["params"]["rho"], simulations[description]["params"]["psi"], R0] + list_params_in_simulation

                sias_params_df.to_csv(simulation_sias_file, sep=";", header = True, mode = 'w', index=False)
            else:
                for param_name, param_value in simulations[description]["params"].items():
                    if (param_value is None):
                        simulations[description]["params"][param_name] = simulation_row[param_name]

                    if (param_name+"_list" in sias_params_df.columns) and (not simulation_row[param_name+"_list"] is None) and isinstance(simulation_row[param_name+"_list"], list):
                        simulations[description]["params_list"][param_name] = simulation_row[param_name+"_list"]

                simulations[description]["params"]["psi"] = simulation_row["psi"]

            print(description + " parameters: ", simulations[description]["params"], "\n")
            transmition_factor_fn = get_transmition_factor_fn(psi=simulations[description]["params"]["psi"], date_i=date_of_effectiveness_of_policy) if (consider_policy_effect) else None
            simulations[description]["results"] = solve_ivp(fun=partial(diff_SIAS_with_p, **{param_key: param_value for param_key, param_value in simulations[description]["params"].items() if (not param_key in ["psi"])}, N=df.iloc[1]["N"], transmition_factor_fn=transmition_factor_fn), t_span=(1, df["t"].max()+150), y0=[S0-p_IA0, p_IA0, I0, 0, 0, 1], method="RK45", max_step=0.1)

            #sias_model_without_sd = get_AS_description(cut_date=date_of_effectiveness_of_policy, p_IA0=p_IA0, policy_begin_date=None, rho=None)
            #params = get_SIAS_params(simulations, sir_key=sir_model_key, sias_key=sias_model_without_sd, psi=psi, use_median=True, policy_date_i=policy_date_i, policy_date_f=policy_date_f)
            #results = solve_ivp(fun=partial(diff_SIAS_with_p, **params, N=data_df.iloc[1]["N"]), t_span=(1, data_df["t"].max()+150), y0=[S0-p_IA0, p_IA0, I0, 0, 0], method="RK45", max_step=0.1)
            #t, IA, IS, D = results.t, results.y[1], results.y[2], results.y[-1]

sias_params_df


# **Table 1**. Estimated values of the epidemiological parameters.

# In[ ]:


chosen_models = collections.OrderedDict()

sias_parameters_df = pd.DataFrame(columns = ["description", "p_IA0", "beta_A", "1/beta_A", "gamma_A", "1/gamma_A", "R_A", "beta_S", "1/beta_S", "gamma_S", "1/gamma_S", "R_S", "rho", "p", "R0"])
for p_IA0 in p_IA0_list:
    cut_date, rho = date_of_effectiveness_of_policy, None
    description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=None, rho=rho)

    beta_A = simulations[description]["params"]["beta_A"]
    beta_S = simulations[description]["params"]["beta_S"]

    gamma_A = simulations[description]["params"]["gamma_A"]
    gamma_S = simulations[description]["params"]["gamma_S"]

    rho = simulations[description]["params"]["rho"]
    p = simulations[description]["params"]["p"]

    R_S = p*beta_S*(1-rho)/gamma_S
    R_A = (1-p)*beta_A/gamma_A

    sias_parameters_df.loc[len(sias_parameters_df)] = [description, p_IA0,
                                                       '{:6.4f}'.format(beta_A) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[description]["params_list"]["beta_A"])),
                                                       '{:6.4f}'.format(1/beta_A) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std([1/v for v in simulations[description]["params_list"]["beta_A"]])),
                                                       '{:6.4f}'.format(gamma_A) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[description]["params_list"]["gamma_A"])),
                                                       '{:6.4f}'.format(1/gamma_A) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std([1/v for v in simulations[description]["params_list"]["gamma_A"]])),
                                                       R_A,
                                                       '{:6.4f}'.format(beta_S) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[sir_model_key]["params_list"]["beta"])),
                                                       '{:6.4f}'.format(1/beta_S) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std([1/v for v in simulations[sir_model_key]["params_list"]["beta"]])),
                                                       '{:6.4f}'.format(gamma_S) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[sir_model_key]["params_list"]["gamma"])),
                                                       '{:6.4f}'.format(1/gamma_S) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std([1/v for v in simulations[sir_model_key]["params_list"]["gamma"]])),
                                                       R_S,
                                                       '{:6.4f}'.format(rho) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[sir_model_key]["params_list"]["rho"])),
                                                       '{:6.4f}'.format(p) + " " + u"\u00B1" + " " + '{:6.4f}'.format(np.std(simulations[description]["params_list"]["p"])),
                                                       R_A+R_S]

#display(HTML(tabulate(sias_parameters_df, headers=sias_parameters_df.columns, tablefmt='html')))
print(tabulate(sias_parameters_df, headers='keys', tablefmt='psql', showindex=False))

cut_date, p_IA0 = date_of_effectiveness_of_policy, 1.0
sias_model_key = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=None, rho=None)

params_sias = dict()
sias_params_df = pd.DataFrame(columns = ["param", "q", "value"])
for q in [0.025, 0.05, 0.5, 0.95, 0.975, None]:
    beta_S_list = simulations[sir_model_key]["params_list"]["beta"]
    gamma_S_list = simulations[sir_model_key]["params_list"]["gamma"]
    rho_list = simulations[sir_model_key]["params_list"]["rho"]

    beta_A_list = simulations[sias_model_key]["params_list"]["beta_A"]
    gamma_A_list = simulations[sias_model_key]["params_list"]["gamma_A"]
    p_list = simulations[sias_model_key]["params_list"]["p"]
    
    if (not q is None):
        key = "quantile-"+str(q)
        f_agg = partial(np.quantile, q=q)
    else:
        key = "mean"
        f_agg = np.mean
    
    params_sias[key] = collections.OrderedDict()
    params_sias[key]["beta_S"] = f_agg(beta_S_list)
    params_sias[key]["gamma_S"] = f_agg(gamma_S_list)
    params_sias[key]["rho"] = f_agg(rho_list)

    params_sias[key]["beta_A"] = f_agg(beta_A_list)
    params_sias[key]["gamma_A"] = f_agg(gamma_A_list)
    params_sias[key]["p"] = f_agg(p_list)

    params_sias[key]["RA"] = f_agg([(1-p)*beta_A/gamma_A for beta_A, gamma_A, p in zip(beta_A_list, gamma_A_list, p_list)])
    params_sias[key]["RS"] = f_agg([p*beta_S*(1-rho)/gamma_S for beta_S, gamma_S, rho in zip(beta_S_list, gamma_S_list, rho_list) for p in p_list])
    params_sias[key]["R0"] = f_agg([(1-p)*beta_A/gamma_A+p*beta_S*(1-rho)/gamma_S for beta_S, gamma_S, rho in zip(beta_S_list, gamma_S_list, rho_list) for beta_A, gamma_A, p in zip(beta_A_list, gamma_A_list, p_list)])

    for param, value in params_sias[key].items():
        sias_params_df.loc[len(sias_params_df)] = [param, q if (not q is None) else "mean", value]

sias_params_df.set_index(["param", "q"]).unstack()
print(sias_params_df.set_index(["param", "q"]).unstack())


# Next, we generate bounds that will be used on the rest of the results.

# In[ ]:


cut_date = date(2020, 3, 30)

bounds_SIAS_filename = os.path.join(output_dir, "bounds_SIAS.pickle")
if (not os.path.exists(bounds_SIAS_filename)):

    bounds_SIAS = collections.OrderedDict()
    for p_IA0 in p_IA0_list:
        print("IA0 = ", p_IA0)
        description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)

        t_bounds, bounds_l, bounds_u = get_bounds_SIAS(data_df, simulations, simulation_key=description, p_IA0=p_IA0, sir_parent_key=sir_model_key, sias_parent_key=get_AS_description(cut_date=date_of_effectiveness_of_policy, p_IA0=p_IA0, policy_begin_date=None, rho=None))

        sias_parent_key = get_AS_description(cut_date=date_of_effectiveness_of_policy, p_IA0=p_IA0, policy_begin_date=None, rho=None)
        params = {"beta_S": f_agg(simulations[sir_model_key]["params_list"]["beta"]),
                  "gamma_S": f_agg(simulations[sir_model_key]["params_list"]["gamma"]),
                  "rho": f_agg(simulations[sir_model_key]["params_list"]["rho"]),
                  "beta_A": f_agg(simulations[sias_parent_key]["params_list"]["beta_A"]),
                  "gamma_A": f_agg(simulations[sias_parent_key]["params_list"]["gamma_A"]),
                  "p": f_agg(simulations[sias_parent_key]["params_list"]["p"])}
        t_steps_ahead = 150
        transmition_factor_fn = get_transmition_factor_fn(psi=simulations[description]["params"]["psi"], date_i=date_of_effectiveness_of_policy)
        results = solve_ivp(fun=partial(diff_SIAS_with_p, **params, mu=0, Lambda=0, N=data_df.iloc[1]["N"], transmition_factor_fn=transmition_factor_fn), t_span=(1, df["t"].max()+t_steps_ahead), y0=[S0-p_IA0, p_IA0, I0, 0, 0, 1], method="RK45", max_step=0.1)
        IA, IS, RA, RS, D = results.y[1:]

        bounds_SIAS[p_IA0] = dict()
        bounds_SIAS[p_IA0]["bounds"] = {"t": t_bounds, "lower": bounds_l, "upper": bounds_u}
        bounds_SIAS[p_IA0]["simulation"] = {"t": results.t, "I_A": IA, "I_S": IS, "R_A": RA, "R_S": RS, "D": D}
    
    with open(bounds_SIAS_filename, 'wb') as f:
        pickle.dump(bounds_SIAS, f, pickle.HIGHEST_PROTOCOL)
else:
    with open(bounds_SIAS_filename, 'rb') as f:
        bounds_SIAS = pickle.load(f)


# **Figure 3**. Short term forecast of the SIASD model taking into account government social distance measures. The solid line corresponds to the last date which the model was estimated, and the dashed line are model predictions. We show the evolution of the infected (assymptomatic, symptomatic and both) with 95% confidence interval. We represent the real data as points.

# In[ ]:


if (generate_figures):
    import matplotlib
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)
    
    p_IA0 = 1.0

    # https://htmlcolorcodes.com/color-names/
    colors_SIAS = {"assymptomatic": "MediumSeaGreen", "symptomatic": "OrangeRed", "both": "Black"}
    colors_SIAS2 = {"assymptomatic": "PaleGreen", "symptomatic": "PeachPuff", "both": "DarkGray"}
    
    fig_num = 1
    num_steps_ahead = 1
    for aggregation_type in ["cumulative"]: # "t", 
        for x_limits in [(data_df["t"].min(), last_t+num_steps_ahead), (22, last_t+num_steps_ahead)]:
            fig, ax = plt.subplots(nrows=1, ncols=1)
    
            if (aggregation_type == "cumulative"):
                line_I_real, = ax.plot(data_df["t"], data_df["cases"], label="Real", color="DarkOrange", marker="*", linestyle="")
                I_max = max(data_df["cases"])
            else:
                I_max = None
    
            for k, (cut_date, colors_dict) in enumerate([(policy_begin_date, colors_SIAS2), (date(2020, 3, 30), colors_SIAS)]):
                description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)
                t = simulations[description]["results"].t
                IA, IS, RA, RS, D = simulations[description]["results"].y[1:]
    
                if (aggregation_type=="cumulative"):
                    I_simulation = np.array(IS)+np.array(RS)+np.array(D)
    
                    color = colors_dict["both"]
                    lower_bound = np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S"])+np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["R_S"])+np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["D"])
                    upper_bound = np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S"])+np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["R_S"])+np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["D"])
                    ax.fill_between(bounds_SIAS[p_IA0]["bounds"]["t"], upper_bound, lower_bound, color=color, alpha=0.1)
                    #ax.fill_between(bounds_SIAS[p_IA0]["bounds"]["t"], bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S+R_S+D"], bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S+R_S+D"], color=color, alpha=0.2)
    
                    t_cut = date2t(cut_date)
                    t_training_indexes = [i for i, ti in enumerate(t) if (ti <= t_cut)]
                    ax.plot(t[t_training_indexes], I_simulation[t_training_indexes], label="Cumulative infected", linestyle="-", color=color)
    
                    t_prediction_indexes = [i for i, ti in enumerate(t) if (ti > t_cut) and (ti <= last_t+num_steps_ahead)]
                    ax.plot(t[t_prediction_indexes], I_simulation[t_prediction_indexes], linestyle="--", color=color)
    
                    if (I_max is None) or ((not I_max is None) and (max(I_simulation[t_prediction_indexes]) > I_max)):
                        I_max = max(I_simulation[t_prediction_indexes])
    
                else:
                    x_asymptomatic = np.array(IA)
                    x_symptomatic = np.array(IS)
                    x_both = np.array(IA)+np.array(IS)
    
                    for I_simulation, I_name in [(IA, "Assymptomatic"), (IS, "Symptomatic"), (IA+IS, "Both")]:
                        color = colors_dict[I_name.lower()]
                        
                        if (cut_date == date(2020, 3, 30)):
                            if (I_name.lower()=="assymptomatic"):
                                ax.fill_between(bounds_SIAS[p_IA0]["bounds"]["t"], bounds_SIAS[p_IA0]["bounds"]["upper"]["I_A"], bounds_SIAS[p_IA0]["bounds"]["lower"]["I_A"], color=color, alpha=0.2)
                            elif (I_name.lower()=="symptomatic"):
                                ax.fill_between(bounds_SIAS[p_IA0]["bounds"]["t"], bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S"], bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S"], color=color, alpha=0.2)
                        else:
                            I_name = None
            
                        t_cut = date2t(cut_date)
                        t_training_indexes = [i for i, ti in enumerate(t) if (ti <= t_cut)]
                        ax.plot(t[t_training_indexes], I_simulation[t_training_indexes], label=I_name, linestyle="-", color=color)
            
                        t_prediction_indexes = [i for i, ti in enumerate(t) if (ti > t_cut) and (ti <= last_t+num_steps_ahead)]
                        ax.plot(t[t_prediction_indexes], I_simulation[t_prediction_indexes], linestyle="--", color=color)
        
                        if (I_max is None) or ((not I_max is None) and (max(I_simulation[t_prediction_indexes]) > I_max)):
                            I_max = max(I_simulation[t_prediction_indexes])
        
            ax.set_xlim(*x_limits)
            ax.set_ylim(0, min(I_max, 8000))
            ax.set_ylabel(("Cumulative " if (aggregation_type=="cumulative") else "") + "Infected")
    
            if (aggregation_type=="t"):
                ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1)
            ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
            ax.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))
            plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
        
            plt.savefig(output_dir+"fig_AS_predictions_"+aggregation_type+"_"+str(fig_num)+".png", dpi=300, bbox_inches='tight')
            plt.show()
            
            fig_num += 1

psi_df = pd.DataFrame(columns = ["t", "date", "psi", "R"])
for k in np.arange(start=1, stop=(date(2020, 3, 30) - date_of_effectiveness_of_policy).days+1):
    cut_date = date_of_effectiveness_of_policy + timedelta(days=int(k))
    description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)
    
    psi_k = simulations[description]["params"]["psi"]
    p_k = simulations[description]["params"]["p"]
    beta_A_k = simulations[description]["params"]["beta_A"]
    beta_S_k = simulations[description]["params"]["beta_S"]
    gamma_A_k = simulations[description]["params"]["gamma_A"]
    gamma_S_k = simulations[description]["params"]["gamma_S"]
    rho_k = simulations[description]["params"]["rho"]

    RA_k = (1-p_k)*psi_k*beta_A_k/gamma_A_k
    RS_k = p_k*psi_k*beta_S_k*(1-rho_k)/gamma_S_k
    psi_df.loc[len(psi_df)] = [date2t(cut_date), cut_date, simulations[description]["params"]["psi"], RA_k+RS_k]

#display(HTML(tabulate(psi_df, headers=psi_df.columns, tablefmt='html')))
print(tabulate(psi_df, headers='keys', tablefmt='psql', showindex=False))

if (generate_figures):
    from matplotlib import rc
    rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    rc('text', usetex=True)
    
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(psi_df["t"], psi_df["psi"], linestyle="-", marker="o")
    ax.set_xticks(psi_df["t"].values)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    ax.set_ylabel(r"$\mathbf{\psi}$", fontsize=20)
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    plt.savefig(output_dir+"fig_AS_psi.png", dpi=300, bbox_inches='tight')
    plt.show()
    
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)


# **Figure 6**. Proportions of asymptomatic and symptomatic over time using I_{A,0} = 0.5. Approximately 38% are asymptomatic in
# March 30, 2020.

# In[ ]:


if (generate_figures):
    cut_date = date(2020, 3, 30)
    p_IA0 = 1.0
    description = get_AS_description(cut_date=cut_date, p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)
    
    num_steps_ahead = 33
    for aggregation_type in ["t", "cumulative"]:
        
        t_bounds = bounds_SIAS[p_IA0]["bounds"]["t"]
        if (aggregation_type == "cumulative"):
            t = bounds_SIAS[p_IA0]["simulation"]["t"]
            IA = list(np.array(bounds_SIAS[p_IA0]["simulation"]["I_A"])+np.array(bounds_SIAS[p_IA0]["simulation"]["R_A"]))
            IS = list(np.array(bounds_SIAS[p_IA0]["simulation"]["I_S"])+np.array(bounds_SIAS[p_IA0]["simulation"]["R_S"])+np.array(bounds_SIAS[p_IA0]["simulation"]["D"]))
    
            IA_l = np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["I_A+R_A"])
            IA_u = np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["I_A+R_A"])
            IS_l = np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S+R_S+D"])
            IS_u = np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S+R_S+D"])
        else:
            t, IA, IS = bounds_SIAS[p_IA0]["simulation"]["t"], bounds_SIAS[p_IA0]["simulation"]["I_A"], bounds_SIAS[p_IA0]["simulation"]["I_S"]
            
            IA_l = np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["I_A"])
            IA_u = np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["I_A"])
            IS_l = np.array(bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S"])
            IS_u = np.array(bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S"])
        IA = np.array(IA)
        IS = np.array(IS)
        IT = IA + IS
        
        proportion_IS_l, proportion_IS_u = IS_l/(IS_l+IA_u), IS_u/(IS_u+IA_l)
        proportion_IA_l, proportion_IA_u = IA_l/(IA_l+IS_u), IA_u/(IA_u+IS_l)
    
        fig, ax1 = plt.subplots(nrows=1, ncols=1)
        #ax1.fill_between(t_bounds, 100*proportion_IS_u, 100*proportion_IS_l, color=colors_SIAS["symptomatic"], alpha=0.2)
        
        ax1.plot(t, 100*IS/IT, label="Symptomatic", color=colors_SIAS["symptomatic"])
        ax1.set_xlim(min(t), min(t)+num_steps_ahead)
        
        if (aggregation_type=="t"):
            ax1.set_ylim(25, 55)
        else:
            ax1.set_ylim(30, 70)
    
        ax1.tick_params(axis='y', labelcolor=colors_SIAS["symptomatic"])
        ax1.yaxis.label.set_color(colors_SIAS["symptomatic"])
        ax1.set_ylabel(("Cumulative " if (aggregation_type=="cumulative") else "") + "Infected percentage (Symptomatic)")
    
        ax2 = ax1.twinx()
        #ax2.fill_between(t_bounds, 100*proportion_IA_u, 100*proportion_IA_l, color=colors_SIAS["assymptomatic"], alpha=0.2)
        ax2.plot(t, 100*IA/IT, label="Assymptomatic", color=colors_SIAS["assymptomatic"])
        
        ax2.set_xlim(min(t), min(t)+num_steps_ahead)
        if (aggregation_type=="t"):
            ax2.set_ylim(45, 75)
        else:
            ax2.set_ylim(30, 70)
    
        ax2.tick_params(axis='y', labelcolor=colors_SIAS["assymptomatic"])
        ax2.yaxis.label.set_color(colors_SIAS["assymptomatic"])
        ax2.set_ylabel(("Cumulative " if (aggregation_type=="cumulative") else "") + "Infected percentage (Assymptomatic)")
    
        ax1.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
        plt.setp(ax1.xaxis.get_majorticklabels(), rotation=45)
        plt.savefig(output_dir+"fig_AS_proportion_"+aggregation_type+".png", dpi=300, bbox_inches='tight')
    
        plt.show()
    
        # Showing the data
        t_disc, IA_disc, IS_disc = cont2disc(t, IA, IS)
        IT_disc = IA_disc + IS_disc
        IA_disc = 100*IA_disc/IT_disc
        IS_disc = 100*IS_disc/IT_disc
        
        proportion_df = pd.DataFrame(data = np.hstack((t_disc.reshape(-1,1), IA_disc.reshape(-1,1), IS_disc.reshape(-1,1))), columns=["t", "IA_percentage", "IS_percentage"])
        proportion_df["date"] = proportion_df["t"].apply(lambda ti: t2date(ti))
        
        #display(HTML(tabulate(proportion_df, headers=proportion_df.columns, tablefmt='html')))
        #print(tabulate(proportion_df, headers='keys', tablefmt='psql', showindex=False))


# **Figure 7**. The effect of the ratio between asymptomatic and symptomatic in March 30, 2020 in the proportion of symptomatic
# in the peak.

# In[ ]:


if (generate_figures):
    rho_from_statistics = 0.032
    
    date_to_calculate_percentage = date(2020, 3, 30)
    t_to_calculate_percentage = date2t(date_to_calculate_percentage)
    
    proportion_df = pd.DataFrame(columns = ["p_IA0", "p", "1/p-1", "p_A/p_S", "peak_T", "peak_S", "peak_S_l", "peak_S_u", "new_rho"])
    for p_IA0 in [v for v in p_IA0_list if (v >= 0.05) and (v <= 4)]:
        
        if (p_IA0 in bounds_SIAS):
            key = get_AS_description(cut_date=date(2020, 3, 30), p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)
            results = simulations[key]["results"]
    
            t, IA, IS = bounds_SIAS[p_IA0]["simulation"]["t"], bounds_SIAS[p_IA0]["simulation"]["I_A"], bounds_SIAS[p_IA0]["simulation"]["I_S"]
            IA_in_date = next(IA_i for ti, IA_i in zip(*cont2disc(t,IA)) if np.isclose(ti, t_to_calculate_percentage))
            IS_in_date = next(IS_i for ti, IS_i in zip(*cont2disc(t,IS)) if np.isclose(ti, t_to_calculate_percentage))
    
            proportion_df.loc[len(proportion_df)] = [p_IA0, simulations[key]["params"]["p"], 1/simulations[key]["params"]["p"]-1, 100*IA_in_date/(IA_in_date+IS_in_date), 100*get_SIR_statistics(t, IA+IS)["peak_percent"], 100*get_SIR_statistics(t, IS)["peak_percent"], 100*get_SIR_statistics(t, bounds_SIAS[p_IA0]["bounds"]["lower"]["I_S"])["peak_percent"], 100*get_SIR_statistics(t, bounds_SIAS[p_IA0]["bounds"]["upper"]["I_S"])["peak_percent"], IS_in_date*rho_from_statistics/(IA_in_date+IS_in_date) ]
    
    proportion_df.sort_values("p", inplace=True)
    #display(HTML(tabulate(proportion_df, headers=proportion_df.columns, tablefmt='html')))
    print(tabulate(proportion_df, headers='keys', tablefmt='psql', showindex=False))
    
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True)
    proportion_df = proportion_df.loc[proportion_df["p_IA0"] != 0.2]
    
    #ax.plot(proportion_df["p_A/p_S"], proportion_df["peak_T"], label="Both", linestyle="-", marker=".", color="Black")
    ax.fill_between(proportion_df["p"], proportion_df["peak_S_u"], proportion_df["peak_S_l"], color="Blue", alpha=0.2)
    ax.plot(proportion_df["p"], proportion_df["peak_S"], label="Symptomatic", linestyle="-", marker=".", color="Blue")
    
    ax.set_xlabel("Parameter p - Symptomatic percentage (%)")
    ax.set_ylabel("Peak (%) of symptomatic")
    
    #ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1) # , bbox_to_anchor=(1, 0.5)
    
    plt.savefig(output_dir+"fig_AS_models_peak_comparison.png", dpi=300, bbox_inches='tight')
    
    #ax2 = ax.twinx()
    #ax2.plot(proportion_df["p_A/p_S"], 100*proportion_df["new_rho"], label="rho", linestyle="--", color="Red")
    #ax2.set_ylabel("Recalculated rho")
    
    plt.show()


# **Figure 5**. Long term forecasts of number of infected for different scenarios using the SIASD model. Black, blue, yellow and red lines represent scenarios I to IV, respectively. While solid lines represent the symptomatic infected individuals, dashed lines represent total infected individuals.

# In[ ]:


if (generate_figures):
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True)
    colors = ["Black", "Blue", "Orange", "Red"]
    cenarios_df = pd.DataFrame(columns=["description", "peak at (symptomatic)", "%peak (symptomatic)", "peak at (both)", "%peak (both)"])
    
    p_IA0 = 1.0
    sias_model_without_sd = get_AS_description(cut_date=date_of_effectiveness_of_policy, p_IA0=p_IA0, policy_begin_date=None, rho=None)
    sias_model_with_sd = get_AS_description(cut_date=date(2020, 3, 30), p_IA0=p_IA0, policy_begin_date=policy_begin_date, rho=None)
    
    for i, (label, psi, policy_date_i, policy_date_f) in enumerate([("I", None, None, None),
                                                                    ("II", simulations[sias_model_with_sd]["params"]["psi"], date_of_effectiveness_of_policy, None),
                                                                    ("III", simulations[sias_model_with_sd]["params"]["psi"], date_of_effectiveness_of_policy, date_of_effectiveness_of_policy+timedelta(days=60)),
                                                                    ("IV", simulations[sias_model_with_sd]["params"]["psi"], date_of_effectiveness_of_policy, date(2020, 6, 16))]):
        params = get_SIAS_params(simulations, sir_key=sir_model_key, sias_key=sias_model_without_sd, psi=psi, use_median=False, policy_date_i=policy_date_i, policy_date_f=policy_date_f)
        #print(label, "... ", end="", flush=True)
        print(label, "... ", flush=True)
        print(params, flush=True)
        print("psi = ", psi)
        print("policy_date_i = ", policy_date_i)
        print("policy_date_f = ", policy_date_f)
        print("")
    
        #continue
        
        results = solve_ivp(fun=partial(diff_SIAS_with_p, **params, N=data_df.iloc[1]["N"]), t_span=(1, data_df["t"].max()+150), y0=[S0-p_IA0, p_IA0, I0, 0, 0, 1], method="RK45", max_step=0.1)
        t, IA, IS, D = results.t, results.y[1], results.y[2], results.y[-1]
    
        SIR_statistics_symptomatic = get_SIR_statistics(t, IS)
        SIR_statistics_both = get_SIR_statistics(t, IA+IS)
        cenarios_df.loc[len(cenarios_df)] = [label + " (" + colors[i] + ")", SIR_statistics_symptomatic["peak_date"], 100*SIR_statistics_symptomatic["peak_percent"], SIR_statistics_both["peak_date"], 100*SIR_statistics_both["peak_percent"]]
    
        t_bounds, bounds_l, bounds_u = get_bounds_SIAS(data_df, simulations, simulation_key=sias_model_without_sd if (label=="I") else sias_model_with_sd, p_IA0=p_IA0, sir_parent_key=sir_model_key, sias_parent_key=sias_model_without_sd, policy_date_i=policy_date_i, policy_date_f=policy_date_f, t_steps_ahead=120)
        ax.fill_between(t_bounds, bounds_u["I_S"], bounds_l["I_S"], color=colors[i], alpha=0.2)
    
        line_I, = ax.plot(t, IA+IS, color=colors[i], linestyle="--")
        line_IS, = ax.plot(t, IS, color=colors[i], label=label)
        ax.set_ylabel("Infected")
    
    print("", flush=True)
    
    ax.legend(fancybox=True, framealpha=1, shadow=True, borderpad=1, loc='upper right')
    
    ax.yaxis.set_major_formatter(ticker.FuncFormatter(lambda x,pos: '{:,.0f}'.format(x)))
    
    ax.set_xlim(40, 140)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(t2date))
    plt.setp(ax.xaxis.get_majorticklabels(), rotation=45)
    
    plt.savefig(output_dir+"fig_AS_models_comparison.png", dpi=300, bbox_inches='tight')
    
    plt.show()
    
    #display(HTML(tabulate(cenarios_df, headers=cenarios_df.columns, tablefmt='html')))
    print(tabulate(cenarios_df, headers='keys', tablefmt='psql', showindex=False))


# In[ ]:

print("The end!")
